from __future__ import annotations
from typing import TYPE_CHECKING
import os
from typing import Callable
from mama.utils.system import System, console, Color, get_colored_text


if TYPE_CHECKING:
    from ..build_config import BuildConfig


class GenericYocto:
    def __init__(self, name: str, platform_define, config: BuildConfig):
        self.name = name ## e.g. 'imx8mp' or 'oclea' or 'xilinx'
        self.platform_define = platform_define ## e.g. 'IMX8MP' or 'OCLEA' or 'XILINX'
        self.config = config
        self.toolchain_file = ''  ## for Docker based build, this is the aarch64_toolchain.cmake
        self.toolchain_dir = ''
        self.compilers = ''  ## g++, gcc and ld
        self.cc_prefix = '' ## e.g. '{self.compilers}aarch64-poky-linux-'
        self.sdk_path = ''  ## Path to SDK libs root
        self.sysroot_path = ''  ## Path to system libs root
        self.include_paths = []  ## Path to additional include dirs
        self.version = '' ## GCC Version
        self.build_dir = 'yocto' ## Generic yocto build dir, overridden by subclass
        self.host_name = 'aarch64-poky-linux' ## default host name, overridden by subclass
        self.distro_version = (1,0,0) ## distro version tuple, e.g. (1, 0, 0) for version 1.0.0


    def bin(self):
        """ {sdk_path}/sysroots/x86_64-pokysdk-linux/usr/bin/aarch64-poky-linux/ """
        if not self.compilers: self.init_default()
        return self.compilers


    def sdk(self):
        """ {sdk_path}/sysroots/x86_64-pokysdk-linux/ """
        if not self.compilers: self.init_default()
        return self.sdk_path


    def sysroot(self):
        """ {sdk_path}/sysroots/cortexa53-crypto-poky-linux/ """
        if not self.compilers: self.init_default()
        return self.sysroot_path


    # forced includes that should be added to compiler flags as -I paths
    def includes(self):
        """ [ '{sdk_path}/sysroots/cortexa53-crypto-poky-linux/usr/include' ] """
        if not self.compilers: self.init_default()
        return self.include_paths

    def gcc_prefix(self):
        """ e.g. {sdk_path}/usr/bin/aarch64-poky-linux/aarch64-poky-linux- """
        if not self.compilers: self.init_default()
        return self.cc_prefix


    def append_env_path(self, paths, env):
        path = os.getenv(env)
        if path: paths.append(path)


    def init_default(self):
        if not self.compilers:
            self.init_toolchain()


    def init_toolchain(self, toolchain_dir=None, toolchain_file=None):
        raise NotImplementedError('init_toolchain must be implemented by subclass')


    def _yocto_toolchain_init(self, toolchain_dir=None, toolchain_file=None,
                              paths=[], envs=[],
                              compiler_name='usr/bin/aarch64-poky-linux/aarch64-poky-linux-gcc',
                              sdk_name='x86_64-pokysdk-linux',
                              sysroot_name='cortexa53-crypto-poky-linux',
                              default_toolchain='usr/share/cmake/cortexa53-crypto-poky-linux-toolchain.cmake'):
        # TODO: expand support to enable Windows host cross-compilation?
        if not System.linux:
            raise RuntimeError(f'{self.name} only supported on Linux')

        # add fallback define for user configuration e.g. XILINX_SDK_HOME
        if not envs:
            envs = [ f'{self.platform_define}_SDK_HOME' ]
        for env in envs:
            self.append_env_path(paths, env)

        for path in paths:
            # Check for Yocto structure
            yocto_sdkpath = os.path.abspath(f'{path}/sysroots/{sdk_name}')
            yocto_sysroot = os.path.abspath(f'{path}/sysroots/{sysroot_name}')
            yocto_compiler = f'{yocto_sdkpath}/{compiler_name}'

            if self.config.verbose:
                console(f'Checking for {self.name} toolchain in: {yocto_compiler} and {yocto_sysroot}')

            found_compiler = os.path.exists(yocto_compiler)
            found_sysroot = os.path.exists(yocto_sysroot)
            if found_compiler and found_sysroot:
                self.sdk_path     = yocto_sdkpath # e.g. {path}/sysroots/x86_64-pokysdk-linux
                self.sysroot_path = yocto_sysroot # e.g. {path}/sysroots/cortexa53-crypto-poky-linux
                self.toolchain_dir = os.path.abspath(path)

                # if original `toolchain_dir` was chosen, then prefer toolchain_file
                if toolchain_file and path == toolchain_dir:
                    self._set_toolchain_file(toolchain_file)
                else:
                    self._set_toolchain_file(f'{self.toolchain_dir}/{default_toolchain}')

                if self.sdk_path and self.sysroot_path:
                    self.compilers = os.path.dirname(yocto_compiler) + '/' # e.g. f'{self.sdk_path}/usr/bin/aarch64-poky-linux/'
                    # replace -gcc at the end with '-' to get the prefix
                    # e.g '{self.sdk_path}/usr/bin/aarch64-poky-linux/aarch64-poky-linux-
                    self.cc_prefix = self.compilers + os.path.basename(compiler_name).replace('-gcc', '-')
                    self.include_paths = [ f'{self.sysroot_path}/usr/include' ]
                    self.version = self.config.get_gcc_clang_fullversion(yocto_compiler, dumpfullversion=True)
                    break

            # add some helpful debug messages on potentially broken toolchain configurations
            if found_compiler and not found_sysroot:
                if self.config.print:
                    console(f'Found compiler at {yocto_compiler} but sysroot not found at {yocto_sysroot}', color=Color.YELLOW)
            elif not found_compiler and found_sysroot:
                if self.config.print:
                    console(f'Found sysroot at {yocto_sysroot} but compiler not found at {yocto_compiler}', color=Color.YELLOW)

        # fallback
        if not self.toolchain_file and toolchain_file:
            if not self._set_toolchain_file(toolchain_file):
                raise FileNotFoundError(f'Toolchain file not found: {toolchain_file}')

        if self.config.print:
            OK  = get_colored_text('OK', 'green')
            BAD = get_colored_text('NOTFOUND', 'red')
            def get_path_status(path):
                return OK if path and os.path.exists(path) else BAD

            console(f'Yocto {self.name} TOOLS:     {get_path_status(self.compilers)} {self.compilers}')
            console(f'      {self.name} SDK path:  {get_path_status(self.sdk_path)} {self.sdk_path}')
            console(f'      {self.name} sysroot:   {get_path_status(self.sysroot_path)} {self.sysroot_path}')
            console(f'      {self.name} toolchain: {get_path_status(self.toolchain_file)} {self.toolchain_file}')

        if not os.path.exists(self.compilers):
            raise EnvironmentError(f'''No {self.name} toolchain compilers detected! 
    Default search paths: {paths} 
    Define env {envs[0]} with path to {self.name} tools.''')


    def _set_toolchain_file(self, toolchain_file):
        if os.path.exists(toolchain_file):
            self.toolchain_file = toolchain_file
            return True
        else:
            console(f'No toolchain file found at: {toolchain_file}', color=Color.RED)
            return False


    def _add_common_cxx_flags(self, add_flag: Callable[[str,str], None]):
        # e.g -DIMX8MP=1 or -DOCLEA=1 or -DXILINX=1
        add_flag(f'-D{self.platform_define}', '1')
        # define YOCTO_LINUX=1 for all generic yocto embedded platforms
        add_flag(f'-DYOCTO_LINUX', '1')
        for path in self.includes():
            add_flag(f'-I {path}')

    def get_cxx_flags(self, add_flag: Callable[[str,str], None]):
        """ Should be overridden by subclass to add platform-specific cxx flags """
        self._add_common_cxx_flags(add_flag)


    def get_ldflags_with_defaults(self, ldflags: dict):
        """ Populates ldflags dict with default values if not already set by user configuration. """
        defaults:dict = {
             # -Wl,--as-needed is important for embedded builds to avoid unnecessarily linking to libraries that are not actually used,
             #  which can cause bloated binaries and potential runtime issues on resource-constrained devices
            '-Wl,--as-needed': ''
        }
        # only set default ldflags if not already set by user configuration
        for k,v in defaults.items():
            if not k in ldflags:
                ldflags[k] = v
        return ldflags


    def get_cmake_build_opts(self) -> list:
        if self.toolchain_file:
            if self.config.print:
                console(f'Toolchain: {self.toolchain_file}')
            return [
                f'{self.platform_define}=TRUE',
                f'CMAKE_TOOLCHAIN_FILE="{self.toolchain_file}"'
            ]
        # NOTE: CMAKE_C_COMPILER and CMAKE_CXX_COMPILER is already configured
        #       by get_preferred_compiler_paths()
        opt = [
            f'{self.platform_define}=TRUE',
            'CMAKE_SYSTEM_NAME=Linux',
            'CMAKE_SYSTEM_VERSION=1',
            'CMAKE_SYSTEM_PROCESSOR=arm64',
            'CMAKE_SYSROOT='+self.sysroot(),
            'CMAKE_AR='+self.cc_prefix+'ar',
            'CMAKE_READELF='+self.cc_prefix+'readelf',
            'CMAKE_STRIP='+self.cc_prefix+'strip',
            'CMAKE_RANLIB='+self.cc_prefix+'ranlib',
            'CMAKE_FIND_ROOT_PATH_MODE_PROGRAM=NEVER', # Use our definitions for compiler tools
            'CMAKE_FIND_ROOT_PATH_MODE_LIBRARY=ONLY', # Search for libraries and headers in the target directories only
            'CMAKE_FIND_ROOT_PATH_MODE_INCLUDE=ONLY',
            'CMAKE_BUILD_WITH_INSTALL_RPATH=ON',
        ]
        return opt


    def get_gnu_build_env(self, environ: dict = {}):
        sysroot = f'--sysroot={self.sysroot()}'
        environ['LDFLAGS'] = sysroot
        environ['CFLAGS'] = sysroot
        environ['CXXFLAGS'] = sysroot

        environ['CC'] = self.cc_prefix + 'gcc'
        environ['CXX'] = self.cc_prefix + 'g++'
        environ['AR'] = self.cc_prefix + 'ar'
        environ['LD'] = self.cc_prefix + 'ld'
        environ['READELF'] = self.cc_prefix + 'readelf'
        environ['STRIP'] = self.cc_prefix + 'strip'
        environ['RANLIB'] = self.cc_prefix + 'ranlib'
        return environ




from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Callable
from .generic_yocto import GenericYocto


if TYPE_CHECKING:
    from ..build_config import BuildConfig


class Xilinx(GenericYocto):
    BUILD_DIR = 'xilinx' # Constant: Xilinx Zync build dir name

    def __init__(self, config: BuildConfig):
        ## Xilinx Zynq UltraScale+ MPSoC
        super().__init__(Xilinx.BUILD_DIR, Xilinx.BUILD_DIR.upper(), config)
        self.build_dir = Xilinx.BUILD_DIR  # override generic build dir
        self.host_name = 'aarch64-xilinx-linux' # override generic host name


    def init_toolchain(self, toolchain_dir=None, toolchain_file=None):
        paths = []
        if toolchain_dir: paths += [ toolchain_dir ]
        # this is the primary search path for Linux cross-builds:
        paths += [ '/opt/petalinux/toolchain' ]
        # these are generic ones:
        paths += [ 'xilinx-toolchain', 'xilinx-toolchain/toolchain' ]

        compiler = 'usr/bin/aarch64-xilinx-linux/aarch64-xilinx-linux-gcc'

        self._yocto_toolchain_init(toolchain_dir, toolchain_file,
                                   paths=paths,
                                   envs=['XILINX_HOME', 'XILINX_SDK'],
                                   compiler_name=compiler,
                                   sdk_name='x86_64-petalinux-linux',
                                   sysroot_name='cortexa72-cortexa53-xilinx-linux',
                                   default_toolchain='aarch64_xilinx_toolchain.cmake')


    def get_cxx_flags(self, add_flag: Callable[[str,str], None]):
        # Match PetaLinux SDK environment-setup flags
        add_flag('-mcpu', 'cortex-a72.cortex-a53+crc')
        add_flag('-mbranch-protection', 'standard')
        self._add_common_cxx_flags(add_flag)




from __future__ import annotations
from typing import TYPE_CHECKING
import os, zipfile, shutil

from .artifactory import artifactory_archive_name, artifactory_upload_ftp
from .util import get_file_size_str, console, normalized_join
from .papa_deploy import PapaFileInfo

if TYPE_CHECKING:
    from .build_target import BuildTarget


def _append_files_recursive(zip: zipfile.ZipFile, rel_path:str, full_path:str):
    # add each file recursively
    if os.path.isdir(full_path):
        root = os.path.dirname(full_path)
        for full_dir, _, files in os.walk(full_path):
            rel_dir = full_dir[len(root):].lstrip('\\/')
            # write the directory into the zip as well
            zip.write(full_dir, rel_dir)
            for file in files:
                src_file = full_dir + '/' + file
                rel_file = rel_dir + '/' + file
                #print(f'src_file:{src_file} rel_file:{rel_file}')
                zip.write(src_file, rel_file)
    else:
        zip.write(full_path, os.path.basename(full_path))


def papa_upload_to(target:BuildTarget, package_full_path:str):
    """
    - target: Target which was configured and packaged
    - package_full_path: Full path to deployed PAPA package
    """
    package_full_path = package_full_path if package_full_path else target.build_dir()
    papa_file = normalized_join(package_full_path, 'papa.txt')
    if not os.path.exists(papa_file):
        raise RuntimeError(f'BuildTarget {target.name} was not deployed because '\
                           f'{package_full_path} does not exist! '\
                            'Add self.papa_deploy() to mamafile deploy()!')

    config = target.config
    archive_name = artifactory_archive_name(target)
    if not archive_name:
        raise Exception(f'Could not get archive name for target: {target.name}')

    archive_path = target.build_dir(archive_name + '.zip')
    if config.verbose:
        console(f'    archiving {papa_file}\n {"":10}-> {archive_path}')

    # archive needs to be created manually to only include the files in papa.txt
    papa = PapaFileInfo(papa_file)
    # create a zip archive with papa.includes, papa.libs and papa.assets
    temp_archive = archive_path + '.tmp'
    with zipfile.ZipFile(temp_archive, 'w',
                         compression=zipfile.ZIP_DEFLATED,
                         compresslevel=8) as zip:
        if config.verbose: console(f'      root {package_full_path}')
        zip.write(papa_file, 'papa.txt') # always add the main manifest file
        for include in papa.includes:
            rel_path = os.path.relpath(include, package_full_path)
            if config.verbose: console(f'      adding {rel_path} {include}')
            _append_files_recursive(zip, rel_path, include)
        for lib in papa.libs:
            rel_path = os.path.relpath(lib, package_full_path)
            if config.verbose: console(f'      adding {rel_path} {lib}')
            if rel_path.startswith('..'):
                raise Exception(f'lib path {lib} is outside of the package path {package_full_path}')
            zip.write(lib, rel_path)
        for asset in papa.assets:
            rel_path = asset.outpath
            if config.verbose: console(f'      adding {rel_path} {asset}')
            zip.write(asset.srcpath, asset.outpath)

    # move the intermediate archive to the final location
    if os.path.exists(archive_path):
        os.remove(archive_path)
    shutil.move(temp_archive, archive_path)

    if config.print:
        size = os.path.getsize(archive_path)
        console(f'  - PAPA Upload {archive_name}  {get_file_size_str(size)}')

    if artifactory_upload_ftp(target, archive_path):
        if config.verbose:
            console(f'  PAPA Uploaded {os.path.basename(archive_path)}')




class DepSource(object):
    def __init__(self, name:str):
        self.name = name
        self.is_git = False
        self.is_pkg = False
        self.is_src = False


    def get_type_string(self):
        if self.is_git: return "Git"
        if self.is_pkg: return "ART"
        if self.is_src: return "Src"


    def get_papa_string(self):
        raise RuntimeError('get_papa_string() not implemented')


    @staticmethod
    def papa_join(*fields):
        """ Join all given fields for PAPA package serialization """
        strings = []
        for field in fields:
            if field:
                if isinstance(field, list):
                    strings += field
                else:
                    strings.append(field)
            else:
                strings.append('')
        return ",".join(strings)



import os, concurrent.futures
from typing import List

from mama.build_config import BuildConfig
from .build_dependency import BuildDependency
from .util import read_text_from, write_text_to, save_file_if_contents_changed
from .utils.system import Color, console, error


def _get_cmake_path_list(paths):
    pathlist = '' 
    for path in paths: pathlist += f'\n    "{path}"'
    return pathlist


def _get_exported_libs(target):
    filtered = []
    allowed = []
    is_linux_like = False
    if target.android or target.linux or target.raspi or target.mips or target.yocto_linux:
        is_linux_like = True
    elif target.msvc:
        allowed = ['.lib']
    elif target.macos:
        allowed = ['.a', '.dylib', '.bundle']
    elif target.ios:
        allowed = ['.a', '.dylib', '.framework']

    #print(f'{target.name: <16} exported: {target.exported_libs}')
    for lib in target.exported_libs:
        if not lib: continue
        if is_linux_like:
            # for linux-like targets allow .a, .so and versioned .so.1.2.3 files
            if lib.endswith('.a') or lib.endswith('.so') \
                or (str.isdigit(lib[-1]) and '.so.' in lib):
                filtered.append(lib)
        else:
            for ext in allowed:
                if lib.endswith(ext):
                    filtered.append(lib)
    #print(f'{target.name: <16} filtered: {filtered}')
    return filtered


def _get_hierarchical_libs(root: BuildDependency):
    deps = []
    syslibs = []
    def add_deps(dep: BuildDependency):
        nonlocal deps, syslibs
        deps += _get_exported_libs(dep.target)
        syslibs += dep.target.exported_syslibs
        for child in dep.get_children():
            add_deps(child)
    add_deps(root)
    return deps + syslibs


def _get_flattened_deps(root: BuildDependency):
    # deps have to be sorted in [parent] [child] order for Unix linkers
    ordered = []
    def add_unique_items(deps: List[BuildDependency]):
        for child in deps:
            if child in ordered: # already in deps list, so we need to move it lower
                ordered.remove(child)
            ordered.append(child)
            add_unique_items(child.get_children())
    add_unique_items(root.get_children())
    return ordered


def get_flat_deps(root: BuildDependency):
    """ Gets flat dependencies, including root """
    return [root] + _get_flattened_deps(root)


def get_deps_that_depend_on_target(root: BuildDependency, target: BuildDependency, deps = []) -> List[BuildDependency]:
    discovered_new = False
    """ Gets all dependencies that depend on the target """
    def depth_first_search_for_target(dep: BuildDependency):
        nonlocal discovered_new, target, deps
        depends = False
        for child in dep.get_children():
            if child in deps:
                continue
            if child == target:
                depends = True
            if depth_first_search_for_target(child):
                deps.append(child)
                depends = True
                discovered_new = True
        return depends
    if depth_first_search_for_target(root) and root not in deps:
        deps.append(root)
        discovered_new = True

    # now that we have the initial deps,
    # we need to further expand it to include second level dependencies
    while discovered_new:
        discovered_new = False
        for d in deps:
            depth_first_search_for_target(d)
            if discovered_new:
                break # restart the outer loop
    return deps


def _get_mama_dependencies_cmake(root: BuildDependency, build:str):
    if not root.get_children():
        return ''
    return f'include("{root.dep_dir}/{build}/mama-dependencies.cmake")'


def _mama_cmake_path(root: BuildDependency):
    if not root.src_dir: # for artifactory pkgs, there is no src_dir
        return f'{root.build_dir}/mama.cmake'
    return f'{root.src_dir}/mama.cmake'


def _save_mama_cmake_and_dependencies_cmake(root: BuildDependency):
    # save the {build}/mama-dependencies.cmake
    _save_dependencies_cmake(root)
    # the following is the proxy `mysource/mama.cmake` file
    # which will reference each mama-dependencies.cmake depending on platform
    _save_mama_cmake(root)


def _get_compile_commands_path(dep: BuildDependency):
    src_build_cmds = f'{dep.src_dir}/build/compile_commands.json'
    bin_build_cmds = f'{dep.build_dir}/compile_commands.json'

    src_exists = os.path.exists(src_build_cmds)
    bin_exists = os.path.exists(bin_build_cmds)

    # choose the latest one
    if src_exists and bin_exists and os.path.getmtime(src_build_cmds) > os.path.getmtime(bin_build_cmds):
        # for src_dir paths we use `${workspaceFolder}` macro:
        return '${workspaceFolder}/build/compile_commands.json'
    if bin_exists:
        # for build dir paths, check if build dir is relative to src dir
        if dep.build_dir.startswith(dep.src_dir):
            # if so, we chop off the src dir and use `${workspaceFolder}/`
            rel_build_dir = f'${{workspaceFolder}}{dep.build_dir[len(dep.src_dir):]}/compile_commands.json'
            return rel_build_dir
        return bin_build_cmds # absolute path for build dir paths
    return None


def _find_matching_platform_config(dep: BuildDependency, configurations):
    config_name = dep.config.name()
    config_arch = dep.config.arch

    # first look for Platform + Arch match such as Windows x64
    for conf in configurations:
        name = str(conf["name"]).lower()
        if config_name in name and config_arch in name:
            return conf

    # then look for only Platform match like 'windows'
    for conf in configurations:
        name = str(conf["name"]).lower()
        if config_name in name:
            return conf
    return None


def _save_vscode_compile_commands(dep: BuildDependency):
    if not dep.src_dir: # for artifactory pkgs, there is no src_dir
        return
    if not dep.is_root:
        return

    cpp_props_path = f'{dep.src_dir}/.vscode/c_cpp_properties.json'
    if not os.path.exists(cpp_props_path):
        return

    commands_path = _get_compile_commands_path(dep)
    if not commands_path:
        return

    # we have a valid path for compile_commands.json, now link it into c_cpp_properties.json
    cpp_props_text = read_text_from(cpp_props_path)
    import json
    props = json.loads(cpp_props_text)
    configurations = props["configurations"]

    platform_config = _find_matching_platform_config(dep, configurations)

    # make a copy of the first config and rename it to the platform name
    if not platform_config and len(configurations) > 0:
        platform_config = configurations[0].copy()
        platform_config['name'] = f'{dep.config.name()} {dep.config.arch}'
        configurations.append(platform_config)

    # set the compile commands for this platform
    if platform_config:
        platform_config["compileCommands"] = commands_path

    new_cpp_props_text = json.dumps(props, indent=4)
    if new_cpp_props_text != cpp_props_text:
        write_text_to(cpp_props_path, new_cpp_props_text)
        if dep.config.print and platform_config:
            console(f'Updated c_cpp_properties.json "{platform_config["name"]}" compileCommands')


def _get_dependency_cmake_defines(dep: BuildDependency):
    name = dep.name
    own_libs = _get_exported_libs(dep.target) + dep.target.exported_syslibs
    all_libs = _get_hierarchical_libs(dep)

    includes = _get_cmake_path_list(dep.target.exported_includes)
    own_libs_list = _get_cmake_path_list(own_libs)
    all_libs_list = _get_cmake_path_list(all_libs)

    # reference name_LIB if it equals name_LIBS
    if own_libs_list == all_libs_list:
        all_libs_list = f'${{{name}_LIB}}'
    #console(f'{name} own_libs: {own_libs_list}')
    #console(f'{name} all_libs: {all_libs_list}')
    return f'${{{name}_INCLUDES}}', \
f'''
# Package {name}
set({name}_INCLUDES {includes})
# only {name} libs
set({name}_LIB {own_libs_list})
# includes {name} libs and all dependency libs
set({name}_LIBS {all_libs_list})
'''


def _save_dependencies_cmake(root: BuildDependency):
    if not root.build_dir_exists():
        return # probably CLEAN, so nothing to save
    outfile = f'{root.build_dir}/mama-dependencies.cmake'
    text = \
'''
# This file is auto-generated by mama build. Do not modify by hand!
'''
    includes_def, package_text = _get_dependency_cmake_defines(root)
    includes_defs = [includes_def]
    text += package_text

    root.flattened_deps = _get_flattened_deps(root)
    for dep in root.flattened_deps:
        includes_def, package_text = _get_dependency_cmake_defines(dep)
        includes_defs.append(includes_def)
        text += package_text

    # and finally, set the MAMA_INCLUDES and MAMA_LIBS
    includes = ' '.join(includes_defs)
    libs = f'${{{root.name}_LIBS}}' # use the root package to get the full flat list of deps
    text += \
f'''
set(MAMA_INCLUDES ${{MAMA_INCLUDES}} {includes})
set(MAMA_LIBS     ${{MAMA_LIBS}}     {libs})
'''

    save_file_if_contents_changed(outfile, text)


def _save_mama_cmake(root: BuildDependency):
    # note: we save verbose include directives, 
    #       because CLion has a hard time detecting macro paths
    c:BuildConfig = root.config

    # gets the defines for a single platform and architecture
    def get_build_dir_defines(build_dir):
        return f'''set(MAMA_BUILD "{build_dir}")
        {_get_mama_dependencies_cmake(root, build_dir)}'''

    text = f'''# This file is auto-generated by mama build. Do not modify by hand!
if(CMAKE_CXX_COMPILER_ID MATCHES "Clang")
    set(CLANG TRUE)
elseif(CMAKE_CXX_COMPILER_ID MATCHES "GNU")
    set(GCC TRUE)
endif()

if(CMAKE_GENERATOR_PLATFORM)
    set(MAMA_CMAKE_ARCH ${{CMAKE_GENERATOR_PLATFORM}})
elseif(ANDROID OR ANDROID_NDK)
    set(MAMA_CMAKE_ARCH ${{ANDROID_ARCH}})
elseif(CMAKE_SYSTEM_PROCESSOR)
    set(MAMA_CMAKE_ARCH ${{CMAKE_SYSTEM_PROCESSOR}})
else()
    message(FATAL_ERROR "MAMA: Missing CMake target architecture!")
endif()

# Initializes the INCLUDE and LIBS, they will overwritten in mama-dependencies.cmake
set(MAMA_INCLUDE "")
set(MAMA_LIBS "")

# Set MAMA_INCLUDES and MAMA_LIBS for each platform
if(ANDROID OR ANDROID_NDK)
    if(MAMA_CMAKE_ARCH MATCHES "(arm64)|(ARM64)")
        set(MAMA_ARCH_ARM64 TRUE)
        {get_build_dir_defines(c.build_dir_android64())}
    else()
        set(MAMA_ARCH_ARM32 TRUE)
        {get_build_dir_defines(c.build_dir_android32())}
    endif()
elseif(WIN32)
    if(MAMA_CMAKE_ARCH MATCHES "(amd64)|(AMD64)|(IA64)|(x64)|(X64)|(x86_64)|(X86_64)")
        set(MAMA_ARCH_X64 TRUE)
        {get_build_dir_defines(c.build_dir_win64())}
    elseif(MAMA_CMAKE_ARCH MATCHES "(X86)|(x86)|(i386)|(i686)")
        set(MAMA_ARCH_X86 TRUE)
        {get_build_dir_defines(c.build_dir_win32())}
    elseif(MAMA_CMAKE_ARCH MATCHES "ARM64")
        set(MAMA_ARCH_ARM64 TRUE)
        {get_build_dir_defines(c.build_dir_winarm64())}
    elseif(MAMA_CMAKE_ARCH MATCHES "ARM")
        set(MAMA_ARCH_ARM32 TRUE)
        {get_build_dir_defines(c.build_dir_winarm32())}
    else()
        message(FATAL_ERROR "MAMA: Unrecognized target architecture '${{MAMA_CMAKE_ARCH}}'")
    endif()
elseif(APPLE)
  if(IOS_PLATFORM)
        set(IOS TRUE)
        set(MAMA_ARCH_ARM64 TRUE) # Always arm64
        {get_build_dir_defines(c.build_dir_ios())}
  else()
    set(MACOS TRUE)
    if(MAMA_CMAKE_ARCH MATCHES "x86_64") # (older x64)
        set(MAMA_ARCH_X64 TRUE)
        {get_build_dir_defines(c.build_dir_macos64())}
    elseif(MAMA_CMAKE_ARCH MATCHES "(arm64)|(ARM64)") # (M1 and later)
        set(MAMA_ARCH_ARM64 TRUE)
        {get_build_dir_defines(c.build_dir_macosarm64())}
    else()
        message(FATAL_ERROR "MAMA: Unrecognized macOS architecture '${{MAMA_CMAKE_ARCH}}'")
    endif()
  endif()
elseif(RASPI)
        set(MAMA_ARCH_ARM32 TRUE)
        {get_build_dir_defines(c.build_dir_raspi32())}
elseif(OCLEA)
        set(MAMA_ARCH_ARM64 TRUE)
        set(YOCTO_LINUX ON)
        add_compile_definitions(OCLEA=1)
        {get_build_dir_defines(c.build_dir_oclea64())}
elseif(XILINX)
        set(MAMA_ARCH_ARM64 TRUE)
        set(YOCTO_LINUX ON)
        add_compile_definitions(XILINX=1)
        {get_build_dir_defines(c.build_dir_xilinx64())}
elseif(IMX8MP)
        set(MAMA_ARCH_ARM64 TRUE)
        set(YOCTO_LINUX ON)
        add_compile_definitions(IMX8MP=1)
        {get_build_dir_defines(c.build_dir_imx8mp())}
elseif(MIPS)
        set(MAMA_ARCH_MIPS TRUE)
        add_compile_definitions(MIPS=1)
        {get_build_dir_defines(c.build_dir_mips())}
elseif(UNIX)
    set(LINUX TRUE)
    if(MAMA_CMAKE_ARCH MATCHES "(amd64)|(AMD64)|(IA64)|(x86_64)")
        set(MAMA_ARCH_X64 TRUE)
        {get_build_dir_defines(c.build_dir_linux64())}
    elseif(MAMA_CMAKE_ARCH MATCHES "(X86)|(x86)|(i386)|(i686)")
        set(MAMA_ARCH_X86 TRUE)
        {get_build_dir_defines(c.build_dir_linux32())}
    elseif(MAMA_CMAKE_ARCH MATCHES "(aarch64)|(AARCH64)|(arm64)|(ARM64)")
        set(MAMA_ARCH_ARM64 TRUE)
        {get_build_dir_defines(c.build_dir_linuxarm64())}
    else()
        message(FATAL_ERROR "MAMA: Unrecognized Linux architecture '${{MAMA_CMAKE_ARCH}}'")
    endif()
else()
    message(FATAL_ERROR "mama build: Unsupported Platform! '${{MAMA_CMAKE_ARCH}}'")
endif()

# Overrides linkage on MSVC to non-debug run-time library (TODO: make this configurable)
if(MSVC)
    add_definitions(-D_ITERATOR_DEBUG_LEVEL=0)
    foreach(MODE "_DEBUG" "_MINSIZEREL" "_RELEASE" "_RELWITHDEBINFO")
        string(REPLACE "/MDd" "/MD" TMP "${{CMAKE_C_FLAGS${{MODE}}}}")
        set(CMAKE_C_FLAGS${{MODE}} "${{TMP}}" CACHE STRING "" FORCE)
        string(REPLACE "/MDd" "/MD" TMP "${{CMAKE_CXX_FLAGS${{MODE}}}}")
        set(CMAKE_CXX_FLAGS${{MODE}} "${{TMP}}" CACHE STRING "" FORCE)
    endforeach(MODE)
endif()
'''
    save_file_if_contents_changed(_mama_cmake_path(root), text)


def load_dependency_chain(root: BuildDependency):
    """
    This is main entrypoint for building the dependency chain.
    All dependencies must be resolved at this stage
    """
    with concurrent.futures.ThreadPoolExecutor() as e:
        def load_dependency(dep: BuildDependency):
            if dep.already_loaded:
                return dep.should_rebuild

            changed = dep.load()
            if dep.config.parallel_load:
                futures = []
                for child in dep.get_children():
                    futures.append(e.submit(load_dependency, child))
                for f in futures:
                    changed |= f.result()
            else:
                for child in dep.get_children():
                    changed |= load_dependency(child)

            dep.after_load()
            return changed
        load_dependency(root)


def print_dependencies(root: BuildDependency):
    names = [dep.name for dep in root.flattened_deps]
    dep_names = " ".join(names) if root.flattened_deps else '<none>'
    console(f'  - {root.name} Dependencies:  {dep_names}')

    all_deps = [root] + root.flattened_deps
    libs = []
    for dep in all_deps:
        libs += [(dep.name, 'L', lib) for lib in dep.target.exported_libs]
        libs += [(dep.name, 'S', lib) for lib in dep.target.exported_syslibs]

    if libs:
        console(f'  - {root.name} Exported Libs:')
        for lib in libs:
            console(f'    {lib[0]} [{lib[1]}] {lib[2]}')
    else:
        console(f'  - {root.name} Exported Libs: <none>')


def execute_task_chain(flat_deps_reverse: List[BuildDependency]):
    for dep in flat_deps_reverse:
        if not os.path.exists(_mama_cmake_path(dep)):
            _save_mama_cmake_and_dependencies_cmake(dep) # save a dummy mama.cmake before build

        if dep.config.verbose:
            console(f'  - Execute Tasks: {dep.name}', color=Color.BLUE)

        # validate we're not building twice
        if dep.already_executed:
            error(f"Critical Error: '{dep.name}' executed by child project")
            raise RuntimeError(f"Cyclical Dependency detected for '{dep.name}'")

        # go through all child deps and make sure they executed
        for c in dep.get_children():
            if not c.already_executed:
                error(f"Critical Error: child '{c.name}' has not been executed before executing target '{dep.name}'")
                raise RuntimeError(f"Child target not executed before target which requires it: {c.name}")

        _save_mama_cmake_and_dependencies_cmake(dep)
        dep.target._execute_tasks()

        # saves a helper autocomplete includes txt file to make adding .vscode include paths easier
        _save_vscode_compile_commands(dep)

        if dep.config.verbose and not dep.config.test:
            if dep.is_root_or_config_target():
                print_dependencies(dep)
            # else:
            #     print_dependencies(dep) # TODO: different output for non-root targets


def find_dependency(root: BuildDependency, name: str) -> BuildDependency:
    """ This is mainly used for finding root target or specific command line target """
    if root.name.lower() == name.lower():
        return root
    for dep in root.get_children():
        found = find_dependency(dep, name)
        if found: return found
    return None



import os

class Asset:
    def __init__(self, relpath, fullpath, category):
        """
        Creates an asset. If category is set, then relpath is ignored during deploy
            relpath  -- Relative path to source
            fullpath -- Single full path
            category -- Deployment category
        """
        reldir = os.path.dirname(relpath)
        self.name     = os.path.basename(fullpath)
        self.outpath  = fullpath[fullpath.find(reldir) + len(reldir):].lstrip('\\/')
        self.srcpath  = fullpath

        if category: self.outpath = f'{category}/{self.outpath}'
        else:        self.outpath = f'{reldir}/{self.outpath}'
        #console(f'asset {self.outpath}')

    def __str__(self):  return self.outpath
    def __repr__(self): return self.outpath



from __future__ import annotations
from typing import List, TYPE_CHECKING
import os
from .utils.system import console, System, Color
from .util import normalized_path, glob_with_name_match, glob_with_extensions
from .types.asset import Asset

if TYPE_CHECKING:
    from .build_target import BuildTarget
    from .build_config import BuildConfig

def is_a_static_library(lib: str):
    if not lib: return False
    return lib.endswith('.a') or lib.endswith('.lib')


def is_a_dynamic_library(lib: str):
    if not lib: return False
    return lib.endswith('.dll')    or lib.endswith('.pdb') \
        or lib.endswith('.dylib')  or lib.endswith('.so')  \
        or lib.endswith('.bundle') or lib.endswith('.framework') \
        or lib.endswith('.aar')    or (str.isdigit(lib[-1]) and '.so.' in lib) # allow versioned .so.1.2.3 files


def is_a_library(lib: str):
    return is_a_static_library(lib) or is_a_dynamic_library(lib)


def target_root_path(target: BuildTarget, path: str, build_dir: bool):
    root = target.build_dir() if build_dir else target.source_dir()
    return normalized_path(os.path.join(root, path))


def get_lib_basename(lib: str|tuple):
    if isinstance(lib, tuple):
        return os.path.basename(lib[0])
    elif lib.startswith('-framework '):
        return lib.split(' ', 1)[1]
    else:
        return os.path.basename(lib)


def get_unique_libnames(items: list):
    added = set() # to track already added items
    unique = list() # it must be a list to preserve order of items
    for item in items:
        basename = get_lib_basename(item)
        if not basename in added:
            added.add(basename)
            unique.append(item)
    return unique


def export_include(target: BuildTarget, include_path: str, build_dir: bool):
    include_path = target_root_path(target, include_path, build_dir=build_dir)
    if os.path.exists(include_path):
        #console(f'export_include={include_path}')
        if not include_path in target.exported_includes:
            target.exported_includes.append(include_path)
        return True
    return False


def export_includes(target: BuildTarget, include_paths, build_dir: bool):
    added = False
    if isinstance(include_paths, str):
        return target.export_include(include_paths, build_dir)
    elif isinstance(include_paths, list):
        for include_path in include_paths:
            added |= target.export_include(include_path, build_dir)
    return added


def export_lib(target: BuildTarget, relative_path: str, build_dir: bool):
    path = target_root_path(target, relative_path, build_dir=build_dir)
    if os.path.exists(path):
        target.exported_libs.append(path)
        target.exported_libs = get_unique_libnames(target.exported_libs)
    else:
        console(f'export_lib failed to find: {path}')


def set_export_libs_and_products(target: BuildTarget, libs_and_deps: List[str]):
    """
    Sets target's exported_libs and build_products from previously serialized
    list of libraries and dependencies
    """
    libs_and_deps = cleanup_libs_list(libs_and_deps)
    only_libs = []
    for lib in libs_and_deps:
        if os.path.exists(lib) and is_a_library(lib):
            only_libs.append(lib)
    target.exported_libs = get_unique_libnames(only_libs)
    target.build_products = get_unique_libnames(libs_and_deps)


def cleanup_libs_list(libs: List[str]):
    """Cleans up libs list by removing invalid entries"""
    cleaned = []
    for lib in libs:
        lib = lib.strip()
        if not lib.endswith('.lib.recipe'):
            cleaned.append(lib)
    return cleaned


# NOTE: clean_intermediate_files is a suggestion !
def clean_intermediate_files(target: BuildTarget):
    # never clean root or always_build targets
    if target.dep.always_build or target.dep.is_root:
        return

    config: BuildConfig = target.config
    should_clean = False

    if target.clean_intermediate_files:
        if config.verbose: console('  clean_intermediate [target.clean_intermediate_files]', color=Color.YELLOW)
        should_clean = True
    # always clean the intermediate files if we just did an upload operation
    elif config.upload:
        if config.verbose: console('  clean_intermediate [config.upload]', color=Color.YELLOW)
        should_clean = True
    # do automatic cleaning if we did not do a targeted build -- this was an automatic build from source
    elif (config.build or config.rebuild or config.update) and config.no_specific_target():
        if config.verbose: console('  clean_intermediate [dependency build cleanup]', color=Color.YELLOW)
        should_clean = True

    if not should_clean:
        return

    files_to_clean = glob_with_extensions(target.build_dir(), ['.obj', '.o'])
    if files_to_clean:
        if target.config.print:
            console(f'Cleaning {len(files_to_clean)} intermediate files in {target.build_dir()}', color=Color.YELLOW)
        for file in files_to_clean:
            if os.path.isfile(file):
                os.remove(file)


def export_libs(target: BuildTarget, path, pattern_substrings: List[str], build_dir: bool, order: list):
    root_path = target_root_path(target, path, build_dir=build_dir)
    libs = glob_with_name_match(root_path, pattern_substrings)
    libs = cleanup_libs_list(libs)

    # ignore root_path/deploy
    root_deploy = root_path + '/deploy/'
    libs = [l for l in libs if not l.startswith(root_deploy)]

    target.exported_libs += libs
    target.exported_libs = get_unique_libnames(target.exported_libs)

    # ordering needs to be applied for ALL exported libs, incase they were added in multiple steps
    if order:
        def lib_index(lib):
            for i in range(len(order)):
                if order[i] in lib: return i
            return len(order)  # if this lib name does not match, put it at the end of the list
        def sort_key(lib):
            return lib_index(lib)
        target.exported_libs.sort(key=sort_key)

    return len(target.exported_libs) > 0


def export_asset(target: BuildTarget, asset: str, category=None, build_dir=False):
    full_asset = target_root_path(target, asset, build_dir=build_dir)
    if os.path.exists(full_asset):
        target.exported_assets.append(Asset(asset, full_asset, category))
        return True
    else:
        console(f'export_asset failed to find: {full_asset}')
        return False


def export_assets(target: BuildTarget, assets_path: str, pattern_substrings: list, category=None, build_dir=True):
    assets_path += '/'
    assets = glob_with_name_match(target_root_path(target, assets_path, build_dir=build_dir), pattern_substrings, match_dirs=False)
    if assets:
        for full_asset in assets:
            target.exported_assets.append(Asset(assets_path, full_asset, category))
        return True
    return False


def find_syslib(target: BuildTarget, name: str, apt: bool, required: bool):
    if target.ios or target.macos:
        if not name.startswith('-framework '):
            raise EnvironmentError(f'Expected "-framework name" but got "{name}"')
        return name # '-framework Foundation'
    elif target.linux:
        roots = [ "/usr/lib" ]

        # They may be located in different places
        if 'LD_LIBRARY_PATH' in os.environ:
            roots += os.environ['LD_LIBRARY_PATH'].split(':')

        compiler_dir = 'aarch64-linux-gnu' if System.aarch64 else 'x86_64-linux-gnu'
        for root in roots:
            for candidate in [
                lambda: f'{root}/{compiler_dir}/{name}',
                lambda: f'{root}/{compiler_dir}/lib{name}.so',
                lambda: f'{root}/{compiler_dir}/lib{name}.so.2',
                lambda: f'{root}/{compiler_dir}/lib{name}.a',
                lambda: f'{root}/lib{name}.so',
                lambda: f'{root}/lib{name}.so.2',
                lambda: f'{root}/lib{name}.a' ]:
                if os.path.exists(candidate()):
                    return name # example: we found `libdl.so`, so just return `dl` for the linker
        if not required: return None
        if apt: raise IOError(f'Error {target.name} failed to find REQUIRED SysLib: {name}  Try `sudo apt install {apt}`')
        raise IOError(f'Error {target.name} failed to find REQUIRED SysLib: {name}  Try installing it with apt.')
    else:
        return name # just export it. expect system linker to find it.


def export_syslib(target: BuildTarget, name: str, apt: bool, required: bool):
    """
    - target: The build target where to add the export syslib
    - name: Name of the system library, eg: lzma
    - apt: if true, then apt suggestion is given
    - required: if true, then an exception is thrown if syslib is not found
    """
    try:
        lib = find_syslib(target, name, apt, required)
        if lib:
            #console(f'Exporting syslib: {name}:{lib}')
            target.exported_syslibs.append(lib)
            target.exported_syslibs = get_unique_libnames(target.exported_syslibs)
            return True
    except IOError:
        if target.config.clean:
            # just export it. expect system linker to find it.
            target.exported_syslibs.append(name)
            target.exported_syslibs = get_unique_libnames(target.exported_syslibs)
            return True
        else:
            raise
    return False


def get_lib_basename(syslib: str):
    if syslib.startswith('-framework '):
        return syslib
    return os.path.basename(syslib)


def _reset_syslib_name(syslib: str):
    """ Resets the syslib name from `/usr/lib/x86_64-linux-gnu/liblzma.so` to `lzma` """
    fname = os.path.basename(syslib)
    if fname.startswith('lib'):
        if fname.endswith('.so'):
            return fname[3:-3]  # pop 'lib'(3) from front and '.so'(3) from back
        if fname.endswith('.a'):
            return fname[3:-2]  # pop 'lib' and '.a'
    return fname


def reload_syslibs(target: BuildTarget, syslibs: List[str]):
    reloaded = []
    for syslib in syslibs:
        if syslib.startswith('-framework '):
            reloaded.append(syslib)
        else:
            libname = _reset_syslib_name(syslib)
            lib = find_syslib(target, libname, apt=None, required=False)
            if not lib: lib = syslib # not found, fall back to original syslib
            reloaded.append(lib)
    target.exported_syslibs = reloaded



from __future__ import annotations
import os, sys, tempfile, platform, psutil, shutil
from typing import List, TYPE_CHECKING
from mama.platforms.oclea import Oclea
from mama.platforms.xilinx import Xilinx
from mama.platforms.mips import Mips
from mama.platforms.android import Android
from mama.platforms.imx8mp import Imx8mp
from mama.platforms.generic_yocto import GenericYocto
import mama.util as util
from .utils.system import System, console, Color
from .utils.sub_process import execute, execute_piped

if System.linux:
    import distro

if TYPE_CHECKING:
    from .build_dependency import BuildDependency

###
# Mama Build Configuration is created only once in the root project working directory
# This configuration is then passed down to dependencies
#
class BuildConfig:
    def __init__(self, args):
        # commands
        self.list    = False
        self.build   = False
        self.clean   = False
        self.rebuild = False
        self.update  = False
        self.deploy  = False
        # if root mamafile has defined an artifacts URL
        # this will upload deploy archive through SFTP
        self.upload  = False
        # currently for uploads only, uploads only if package already not uploaded
        self.if_needed = False
        # if `art` is specified, then artifactory download is mandatory, no source builds are done
        self.force_artifactory = False
        # if `noart` is specified, then artifactory is temporarily ignored
        self.disable_artifactory = False
        self.reclone   = False
        self.dirty     = False # marks a target for rebuild on next build even if it's up to date
        self.unshallow = False  # by default, git clones are shallow, this allows unshallowing
        self.mama_init = False
        self.print     = True
        self.verbose   = False
        self.test      = ''
        self.start     = ''
        self.with_tests = False # forces -DENABLE_TESTS=ON
        self.test_until_failure = 0 # if > 0, runs test executable in a loop until it fails, useful for catching flaky tests
        self.sanitize  = None # gcc/clang: -fsanitize=[thread|leak|address|undefined]
        self.coverage  = None # gcc/clang: gcov | msvc: /fsanitize-coverage=edge
        self.coverage_report = None # runs gcovr to generate coverage report
        self.clang_tidy_path = None # enables clang-tidy static analysis during build
        # supported platforms
        self.msvc    = False # whether this is a MSVC build on Windows
        self.linux   = False
        self.macos   = False
        self.ios     = False
        self.android : Android = None
        self.raspi   = False # TODO: modernize to Class based impl
        self.mips : Mips = None
        self.oclea : Oclea = None
        self.xilinx : Xilinx = None
        self.imx8mp : Imx8mp = None
        self.yocto_linux : GenericYocto = None # whether this is a generic Yocto Linux embedded platform (e.g. Oclea, Xilinx, IMX8MP)
        # cmake customization
        self.cmake_command = 'cmake' # by default, use whatever cmake is in PATH
        # compiler preferences
        self.clang = False
        self.gcc   = False
        self.clang_path = ''
        self.gcc_path = ''
        # can be used to overide C and C++ compiler paths
        self.cc_path = ''
        self.cxx_path = ''
        self.cxx_version = '' # c++ compiler version, eg '8.3.0' for gcc 8.3.0
        # If compiler specificed from command line
        # using `mama build gcc` or `mama build clang`
        self.compiler_cmd = False
        self.fortran = ''
        # build optimization
        self.release = True
        self.debug   = False
        # valid architectures: x86, x64, arm, arm64
        self.arch    = None
        self.distro  = None  # distro information (name, major, minor)
        self.jobs    = psutil.cpu_count()
        self.target  = None
        self.flags   = None
        self.open    = None
        # use this to customize ios sdk version
        self.ios_version   = '16.0' # 16: ios 16
        # use this to customize macos sdk version
        self.macos_version = '13.0' # 13: macos 13
        ## Artifactory URL for dependency uploads and downloads
        self.artifactory_ftp = None
        self.artifactory_auth = None
        ## Ninja
        self.ninja_path = self.find_ninja_build()
        self.prefer_ninja = not System.windows # do not prefer ninja on Windows by default
        ## MSVC, MSBuild
        self._visualstudio_path = None
        self._visualstudio_cmake_id = None
        self._msbuild_path = None
        self._msvctools_path = None
        ## Raspberry PI - Raspi
        self.raspi_compilers  = ''  ## Raspberry g++ and gcc
        self.raspi_system     = ''  ## path to Raspberry system libraries
        self.raspi_include_paths = [] ## path to additional Raspberry include dirs
        ## Convenient installation utils:
        self.convenient_install = []
        ## Workspace and parsing
        self.parallel_load = False  ## Whether to load dependencies in parallel?
        self.global_workspace = False
        if System.windows:
            self.workspaces_root = util.normalized_path(os.getenv('HOMEPATH'))
        else:
            self.workspaces_root = os.getenv('HOME')
        self.unused_args = []
        self.loaded_dependencies : dict[str, BuildDependency] = {}
        self.parse_args(args)
        self.check_platform()


    def parse_args(self, args: List[str]):
        for arg in args:
            if   arg == 'list':      self.list    = True
            elif arg == 'build':     self.build   = True
            elif arg == 'clean':     self.clean   = True
            elif arg == 'rebuild':   self.rebuild = True
            elif arg == 'update':    self.update  = True
            elif arg == 'deploy':    self.deploy  = True
            elif arg == 'upload':    self.upload  = True
            elif arg == 'if_needed': self.if_needed = True
            elif arg == 'art':       self.force_artifactory = True
            elif arg == 'noart':     self.disable_artifactory = True
            # Updates, Builds and Deploys the project as a package
            elif arg == 'serve':
                self.rebuild = True
                self.update = True
                self.deploy = True
                self.upload = True
            elif arg == 'reclone':
                console('WARNING: Argument `reclone` is deprecated, use `wipe` instead.')
                self.reclone = True
            elif arg == 'wipe':      self.reclone = True
            elif arg == 'dirty':     self.dirty = True
            elif arg == 'unshallow': self.unshallow = True
            elif arg == 'init':      self.mama_init = True
            elif arg == 'silent':    self.print = False
            elif arg == 'verbose':   self.verbose = True
            elif arg == 'parallel':  self.parallel_load = True
            elif arg == 'all':       self.target = 'all'
            elif arg == 'test':      self.test = ' ' # no test arguments
            elif arg == 'start':     self.start = ' ' # no start arguments
            elif arg == 'with_tests': self.with_tests = True
            elif arg.startswith('sanitize='): self.add_sanitizer_option(arg[9:])
            elif arg == 'asan':    self.add_sanitizer_option('address')
            elif arg == 'lsan':    self.add_sanitizer_option('leak')
            elif arg == 'tsan':    self.add_sanitizer_option('thread')
            elif arg == 'ubsan':   self.add_sanitizer_option('undefined')
            elif arg == 'clang-tidy': self.set_clang_tidy_path()
            elif arg.startswith('coverage='): self.add_coverage_option(arg[9:])
            elif arg == 'coverage': self.add_coverage_option()
            elif arg == 'coverage-report':
                self.coverage_report = '.'
                self.add_coverage_option() # also enable coverage if reporting is requested
            elif arg.startswith('coverage-report='):
                self.coverage_report = arg[16:]
                self.add_coverage_option()
            elif arg == 'windows': self.set_platform(msvc=True)
            elif arg == 'msvc':    self.set_platform(msvc=True)
            elif arg == 'linux':   self.set_platform(linux=True)
            elif arg == 'macos':   self.set_platform(macos=True)
            elif arg == 'ios':     self.set_platform(ios=True)
            elif arg == 'android': self.set_platform(android=True)
            elif arg == 'raspi':   self.set_platform(raspi=True)
            elif arg == 'oclea':   self.set_platform(oclea=True)
            elif arg == 'xilinx':  self.set_platform(xilinx=True)
            elif arg == 'mips':    self.set_platform(mips=True)
            elif arg == 'imx8mp':  self.set_platform(imx8mp=True)
            elif arg == 'x86':     self.set_arch('x86')
            elif arg == 'x64':     self.set_arch('x64')
            elif arg == 'arm':     self.set_arch('arm')
            elif arg == 'arm64':   self.set_arch('arm64')
            elif arg == 'aarch64':
                console('warning: aarch64 is the same as arm64, setting to arm64')
                self.set_arch('arm64')
            elif arg == 'clang':
                self.gcc = False
                self.clang = True
                self.compiler_cmd = True
            elif arg == 'gcc':
                self.gcc = True
                self.clang = False
                self.compiler_cmd = True
            elif arg == 'fortran': self.fortran = self.find_default_fortran_compiler()
            elif arg == 'fortran=': self.fortran = arg[8:]
            elif arg == 'release': self.set_build_config(release=True)
            elif arg == 'debug':   self.set_build_config(debug=True)
            elif arg == 'open':    self.open = 'root'
            # Open a specific target source dir for editing with VSCode or Visual Studio
            # Ex old: mama open=ReCpp
            # Ex new: mama open ReCpp
            elif arg.startswith('open='):   self.open = arg[5:]
            elif arg.startswith('jobs='):   self.jobs = int(arg[5:])
            # Sets the target to build/update/clean
            # This is superceded by automatic target lookup
            # Ex old: mama build target=opencv
            # Ex new: mama build opencv
            elif arg.startswith('target='): self.target = arg[7:]
            # Adding arguments for tests runner
            # Ex: mama build test="nogdb threadpool"
            # Ex: mama build test=nogdb test=threadpool
            # Ex: mama build test=nogdb,threadpool
            elif arg.startswith('test='):   self.test = self.join_args(self.test, arg[5:])
            # Adding arguments for test runner to run tests in a loop until failure, useful for catching flaky tests
            # Ex: mama build test="my_flaky_test" test_until_failure=100
            elif arg == 'test_until_failure': self.test_until_failure = 1000 # arbitrary default
            elif arg.startswith('test_until_failure='): self.test_until_failure = int(arg[19:]) # set number of iterations to run tests until failure
            # Calls target.start with the specified arguments
            # Ex: mama build start=verify
            elif arg.startswith('start='):  self.start = self.join_args(self.start, arg[6:])
            elif arg.startswith('arch='):   self.set_arch(arg[5:])
            # Add additional compiler flags
            elif arg.startswith('flags='):  self.flags = self.join_args(self.flags, arg[6:])
            # Ex: mama build android-24
            elif arg.startswith('android-'):
                self.set_platform(android=True)
                self.android.android_api = arg
            elif arg.startswith('install-clang-'): self.convenient_install.append('clang-' + arg[14:])
            elif arg.startswith('install-gcc-'):   self.convenient_install.append('gcc-' + arg[12:])
            elif arg == 'install-msbuild': self.convenient_install.append('msbuild')
            elif arg == 'install-ndk':     self.convenient_install.append('ndk')
            else:
                self.unused_args.append(arg)
            continue


    # modifies existing `args` by parsing and appending `arg` contents
    def join_args(self, args, arg):
        if arg[0] == '"' and arg[-1] == '"':
            arg = arg[1:-1]
        elif ',' in arg:
            arg = ' '.join(arg.split(','))
        if not args:
            return arg
        return args + ' ' + arg


    def set_platform(self, msvc=False, linux=False, macos=False, \
                           ios=False, android=False, raspi=False, \
                           oclea=False, mips=False, xilinx=False, imx8mp=False):
        """ Ensures only a single platform is set """

        platforms = [False]*10
        if   msvc:   platforms[0] = True
        elif linux:  platforms[1] = True
        elif macos:  platforms[2] = True
        elif ios:    platforms[3] = True
        elif android:platforms[4] = True
        elif raspi:  platforms[5] = True
        elif oclea:  platforms[6] = True
        elif mips:   platforms[7] = True
        elif xilinx: platforms[8] = True
        elif imx8mp: platforms[9] = True

        def get_new_value(old_value, enable, type=None):
            if old_value and not enable:
                return None if type else False
            if enable and not old_value:
                return type(self) if type else True
            return old_value
        self.msvc    = get_new_value(self.msvc,    platforms[0])
        self.linux   = get_new_value(self.linux,   platforms[1])
        self.macos   = get_new_value(self.macos,   platforms[2])
        self.ios     = get_new_value(self.ios,     platforms[3])
        self.android = get_new_value(self.android, platforms[4], Android)
        self.raspi   = get_new_value(self.raspi,   platforms[5])
        self.oclea   = get_new_value(self.oclea,   platforms[6], Oclea)
        self.mips    = get_new_value(self.mips,    platforms[7], Mips)
        self.xilinx  = get_new_value(self.xilinx,  platforms[8], Xilinx)
        self.imx8mp  = get_new_value(self.imx8mp,  platforms[9], Imx8mp)

        # convenience alias for detecting embedded Yocto Linux platforms (e.g. Oclea, Xilinx, IMX8MP)
        if self.imx8mp: self.yocto_linux = self.imx8mp
        if self.oclea:  self.yocto_linux = self.oclea
        if self.xilinx: self.yocto_linux = self.xilinx
        return True


    def is_platform_set(self):
        return self.msvc or self.linux or self.macos \
            or self.ios or self.android or self.raspi or self.mips \
            or self.yocto_linux


    def check_platform(self):
        if not self.is_platform_set():
            # choose MSVC if user did not specify `mama build gcc` on windows system
            msvc = System.windows and not (self.gcc or self.clang)
            self.set_platform(msvc=msvc, linux=System.linux, macos=System.macos)
            if not self.is_platform_set():
                raise RuntimeError(f'Unsupported platform {sys.platform}: Please specify platform!')
        if not self.msvc and not (self.gcc or self.clang):
            self.gcc = True # default to GCC on non-MSVC platforms

        # set defaults if arch was not specified
        if not self.arch:
            # macos has now moved to arm64 starting with M1 series chips
            if self.macos:        self.set_arch('arm64')
            elif self.ios:        self.set_arch('arm64')
            elif self.android:    self.set_arch('arm64')
            elif self.yocto_linux:self.set_arch('arm64')
            elif self.raspi:      self.set_arch('arm')
            elif self.mips:       self.set_arch(self.mips.mips_arch)
            else:
                if System.aarch64:  self.set_arch('arm64')
                elif System.x86_64: self.set_arch('x64')
                else:               self.set_arch('x86')

        # Arch itself is validated in set_arch(),
        # however we need to validate if arch is allowed on platform
        if self.arch:
            if self.yocto_linux and self.arch != 'arm64':
                raise RuntimeError(f'Unsupported arch={self.arch} on {self.yocto_linux.name} platform! Supported=arm64')
            if self.linux and self.arch == 'arm':
                raise RuntimeError(f'Unsupported arch={self.arch} on linux platform! Build with android instead')
            if self.raspi and self.arch != 'arm':
                raise RuntimeError(f'Unsupported arch={self.arch} on raspi platform! Supported=arm')
            if self.mips and self.arch not in self.mips.supported_arches:
                raise RuntimeError(f'Unsupported arch={self.arch} on MIPS platform! Supported={self.mips.supported_arches}')


    def get_distro_info(self):
        if self.distro:
            return self.distro
        if self.msvc:
            version = platform.version().split('.') + ['0']
            self.distro = (self.name(), int(version[0]), int(version[1]))
        elif self.macos:
            version = self.macos_version.split('.') + ['0']
            self.distro = (self.name(), int(version[0]), int(version[1]))
        elif self.ios:
            version = self.ios_version.split('.') + ['0']
            self.distro = (self.name(), int(version[0]), int(version[1]))
        elif self.android:
            version = self.android.android_api.split('-')[1]
            self.distro = (self.name(), int(version), 0)
        elif self.yocto_linux:
            self.distro = self.yocto_linux.distro_version
        elif self.raspi:
            # TODO: RASPI version
            self.distro = (self.name(), 0, 0)
        elif self.mips:
            # TODO: MIPS version
            self.distro = (self.name(), self.mips.toolchain_major, self.mips.toolchain_minor)
        elif self.linux:
            try:
                dist = distro.info()
                major = int(dist['version_parts']['major'])
                minor = int(dist['version_parts']['minor'])
                self.distro = (dist['id'], major, minor)
            except Exception as err:
                console(f'Failed to parse linux distro; falling back to Ubuntu 16.04 LTS: {err}', color=Color.RED)
                self.distro = ('ubuntu', 16, 4)
        else:
            self.distro = (platform.system().lower(), int(platform.release()), 0)
        return self.distro


    def set_arch(self, arch):
        arches = ['x86', 'x64', 'arm', 'arm64', 'mips', 'mipsel', 'mips64', 'mips64el']
        if not arch in arches:
            raise RuntimeError(f"Unrecognized architecture {arch}! Valid options are: {arches}")
        self.arch = arch


    def is_64bit_build(self):
        return (self.arch == 'x64' or self.arch == 'arm64')


    def name(self):
        if self.msvc:    return 'windows' # TODO: maybe rename to MSVC?
        if self.linux:   return 'linux'
        if self.macos:   return 'macos'
        if self.ios:     return 'ios'
        if self.android: return 'android'
        if self.raspi:   return 'raspi'
        if self.yocto_linux: return self.yocto_linux.name
        if self.mips:        return self.mips.name
        return 'build'


    ## These are the hard references to all build directory variations
    ## All parts of the codebase should use these, instead of raw strings
    ## This will avoid accidental mismatches
    def build_dir_win64(self): return 'windows'
    def build_dir_win32(self): return 'windows32'
    def build_dir_winarm64(self): return 'winarm'
    def build_dir_winarm32(self): return 'winarm32'
    def build_dir_linux64(self): return 'linux'
    def build_dir_linux32(self): return 'linux32'
    def build_dir_linuxarm64(self): return 'linuxarm' # arm64
    def build_dir_macosarm64(self): return 'macosarm' # arm64
    def build_dir_macos64(self): return 'macos' # x64
    def build_dir_ios(self): return 'ios' # arm64
    def build_dir_android64(self): return 'android'
    def build_dir_android32(self): return 'android32'
    def build_dir_raspi32(self): return 'raspi'
    def build_dir_oclea64(self): return Oclea.BUILD_DIR
    def build_dir_xilinx64(self): return Xilinx.BUILD_DIR
    def build_dir_imx8mp(self): return Imx8mp.BUILD_DIR
    def build_dir_mips(self): return 'mips'
    def build_dir_default(self): return 'build'


    def platform_build_dir_name(self):
        """
        Gets the build folder name depending on platform and architecture.
        By default 64-bit architectures use the platform name, eg 'windows' or 'linux'
        And 32-bit architectures add a suffix, eg 'windows32' or 'linux32'
        """
        # WARNING: This needs to be in sync with dependency_chain.py: _save_mama_cmake !!!
        if self.msvc:
            if self.is_target_arch_x64(): return self.build_dir_win64()
            if self.is_target_arch_x86(): return self.build_dir_win32()
            if self.is_target_arch_armv7(): return self.build_dir_winarm32()
            return self.build_dir_winarm64()
        if self.yocto_linux:
            # Only arm64 i.MX8M Plus, Xilinx Zync, Oclea Aarch64...
            return self.yocto_linux.build_dir
        if self.linux:
            if self.is_target_arch_x64(): return self.build_dir_linux64()
            if self.is_target_arch_arm64(): return self.build_dir_linuxarm64()
            return self.build_dir_linux32()
        if self.macos: # Apple dropped 32-bit support
            # and new default should be arm64 starting from M1 series chips
            if self.is_target_arch_arm64(): return self.build_dir_macosarm64()
            return self.build_dir_macos64()
        if self.ios: # Apple dropped 32-bit support
            return self.build_dir_ios()
        if self.android:
            if self.is_target_arch_arm64(): return self.build_dir_android64()
            return self.build_dir_android32()
        if self.raspi: return self.build_dir_raspi32()  # Only 32-bit raspi
        if self.mips: return self.build_dir_mips()

        return self.build_dir_default()


    def set_build_config(self, release=False, debug=False):
        self.release = release
        self.debug   = debug
        return True


    def set_artifactory_ftp(self, ftp_url: str, auth='store'):
        """ @see BuildTarget.set_artifactory_ftp() for documentation """
        self.artifactory_ftp = ftp_url
        self.artifactory_auth = auth


    def prefer_clang(self, target_name):
        if not self.linux or self.raspi or self.clang: return
        if not self.compiler_cmd:
            self.clang = True
            self.gcc   = False
            self.compiler_cmd = True
            if self.print:
                console(f'Target {target_name} requests Clang. Using Clang since no explicit compiler flag passed.')
        else:
            if self.print:
                console(f'Target {target_name} requested Clang but compiler already set to GCC.')


    def prefer_gcc(self, target_name):
        if not self.linux or self.raspi or self.gcc: return
        if not self.compiler_cmd:
            self.clang = False
            self.gcc   = True
            self.compiler_cmd = True
            if self.print:
                console(f'Target {target_name} requests GCC. Using GCC since no explicit compiler flag passed.')
        else:
            if self.print:
                console(f'Target {target_name} requested GCC but compiler already set to Clang.')


    ##
    # Enables fortran compiler
    # @path Optional custom path or command for the Fortran compiler
    #
    def enable_fortran(self, path=''):
        if self.fortran: return
        self.fortran = path if path else self.find_default_fortran_compiler()


    # returns: root path where the compilers exist and the discovered suffix
    #          (root_path, suffix)
    def find_compiler_root(self, suggested_path, compiler, suffixes, dumpfullversion) -> tuple[str, str, str]:
        def resolve_compiler(cxx_path, suffix) -> tuple[str, str, str]:
            cxx_path = os.path.realpath(cxx_path) # resolve symlinks
            version = self.get_gcc_clang_fullversion(cxx_path, dumpfullversion)
            if self.verbose: console(f'Compiler {cxx_path} version: {version}')
            return os.path.dirname(cxx_path) + '/', suffix, version

        # stop search early if we meet an already pre-configure /etc/alternatives/clang++ path on linux
        # since this is likely what the user has configured as their default compiler
        priority_choices = [ suggested_path, os.getenv('CXX'), '/etc/alternatives/' + compiler ]
        for priority_cxx in priority_choices:
            if priority_cxx and os.path.exists(priority_cxx):
                return resolve_compiler(priority_cxx, '')

        # perform exhaustive search through all candidate directories for any suitable compilers
        roots = []
        if suggested_path: roots.append(suggested_path)
        roots += ['/etc/alternatives/', '/usr/bin/', '/usr/local/bin/', '/bin/']

        # Look in PATH in addition to hardcoded paths
        pathDirs = os.getenv('PATH').split(":")
        pathDirs = list(map(lambda p: p if p.endswith("/") else p + "/", pathDirs)) # Add slash at end if missing
        roots += pathDirs

        candidates = []
        for root in roots:
            for suffix in suffixes:
                cc_path = root + compiler + suffix
                if os.path.exists(cc_path):
                    candidates.append(resolve_compiler(cc_path, suffix))
        if not candidates:
            raise EnvironmentError(f'Could not find {compiler} from {roots} with any suffix {suffixes}')

        def version_to_int(version_str):
            major_minor_patch = version_str.split('.')
            integer = 0
            for part in major_minor_patch:
                integer = integer*10 + int(part) if part else integer
            if integer == 0:
                console(f"Failed to check version for candidate='{version_str}'")
            return integer

        # sort by version, descending eg 10.3, 9.4, 8.3
        candidates.sort(key=lambda x: version_to_int(x[2]), reverse=True)

        # print this out for debugging on CI machines if they select verbose
        if self.verbose:
            for root, suffix, version in candidates:
                console(f'Compiler {compiler+suffix} ({version}) at {root+compiler+suffix}')

        root, suffix, version = candidates[0]
        if self.verbose:
            console(f'==> Selected {compiler+suffix} ({version}) at {root+compiler+suffix} <==')
        return root, suffix, version


    def get_preferred_compiler_paths(self):
        if self.cc_path and self.cxx_path and self.cxx_version:
            return (self.cc_path, self.cxx_path, self.cxx_version)

        # no preferred cc path for MSVC
        if self.msvc:
            return (self.cc_path, self.cxx_path, self.cxx_version)

        if self.android:
            self.cc_path  = self.android.cc_path()
            self.cxx_path = self.android.cxx_path()
            self.cxx_version = self.get_gcc_clang_fullversion(self.cc_path, dumpfullversion=False)
        elif self.yocto_linux:
            self.cc_path  = f'{self.yocto_linux.cc_prefix}gcc'
            self.cxx_path = f'{self.yocto_linux.cc_prefix}g++'
            self.cxx_version = self.get_gcc_clang_fullversion(self.cc_path, dumpfullversion=True)
        elif self.raspi:  # only GCC available for this platform
            ext = '.exe' if System.windows else ''
            self.cc_path  = f'{self.raspi_bin()}arm-linux-gnueabihf-gcc{ext}'
            self.cxx_path = f'{self.raspi_bin()}arm-linux-gnueabihf-g++{ext}'
            self.cxx_version = self.get_gcc_clang_fullversion(self.cc_path, dumpfullversion=True)
        elif self.mips:
            self.cc_path  = f'{self.mips.compiler_prefix()}gcc'
            self.cxx_path = f'{self.mips.compiler_prefix()}g++'
            self.cxx_version = self.get_gcc_clang_fullversion(self.cc_path, dumpfullversion=True)
        elif self.clang:
            suffixes = ['-20','-19','-18','-17','-16','-15','-14','-13','-12','-11','-10','-9','-8','-7','-6','']
            self.clang_path, suffix, ver = self.find_compiler_root(self.clang_path, 'clang++', suffixes, dumpfullversion=False)
            self.cc_path = f'{self.clang_path}clang{suffix}'
            self.cxx_path = f'{self.clang_path}clang++{suffix}'
            self.cxx_version = ver
        elif self.gcc:
            suffixes = ['-15','-14','-13','-12','-11','-10','-9','-8','-7','-6','']
            self.gcc_path, suffix, ver = self.find_compiler_root(self.gcc_path, 'g++', suffixes, dumpfullversion=True)
            self.cc_path = f'{self.gcc_path}gcc{suffix}'
            self.cxx_path = f'{self.gcc_path}g++{suffix}'
            self.cxx_version = ver

        if self.cc_path and self.cxx_path and self.cxx_version:
            return (self.cc_path, self.cxx_path, self.cxx_version)

        raise EnvironmentError('No preferred compiler for this platform!')


    def get_gcc_clang_fullversion(self, cc_path, dumpfullversion):
        if dumpfullversion:
            version = execute_piped([cc_path, '-dumpfullversion']).strip() # eg 9.4.0
            if version.count('.') >= 1:
                return version
        # clang++ doesn't support -dumpfullversion in latest releases -_-
        return execute_piped([cc_path, '-dumpversion']).strip()


    def compiler_version(self):
        if self.msvc:
            msvc_tools = self.get_msvc_tools_path()
            version = os.path.basename(msvc_tools.rstrip('\\//')).split('.')[0]
            return f'msvc{version}'
        elif self.macos:
            return self.macos_version
        elif self.ios:
            return self.ios_version
        elif self.linux or self.raspi or self.mips or self.android or self.yocto_linux:
            cc, _, version = self.get_preferred_compiler_paths()
            version_parts = version.split('.')
            major_version, minor_version = version_parts[0], version_parts[1]
            if 'gcc' in cc: return f'gcc{major_version}.{minor_version}'
            if 'clang' in cc: return f'clang{major_version}.{minor_version}'
            raise EnvironmentError(f'Unrecognized compiler {cc}!')
        else:
            raise EnvironmentError(f'Unknown compiler version!')


    def find_ninja_build(self):
        ninja_executables = [
            os.getenv('NINJA'),
            util.find_executable_from_system('ninja'),
            '/Projects/ninja.exe'
        ]
        for ninja_exe in ninja_executables:
            if ninja_exe and os.path.isfile(ninja_exe):
                if self.verbose: console(f'Found Ninja Build System: {ninja_exe}')
                return ninja_exe
        return ''


    def set_clang_tidy_path(self, clang_tidy_path=None):
        if clang_tidy_path and os.path.exists(clang_tidy_path):
            self.clang_tidy_path = clang_tidy_path
            if self.print: console(f'Using clang-tidy from {clang_tidy_path}', color=Color.GREEN)
            return

        # respect user overrides first
        clang_tidy_env = os.getenv('CLANG_TIDY')
        if clang_tidy_env:
            if os.path.exists(clang_tidy_env):
                self.clang_tidy_path = clang_tidy_env
                if self.print: console(f'Using clang-tidy from CLANG_TIDY env: {clang_tidy_env}', color=Color.GREEN)
                return
            else:
                console(f'CLANG_TIDY environment variable is set to \'{clang_tidy_env}\' but it is not a valid file!', color=Color.YELLOW)

        # display the full path of clang-tidy by resolving symlinks (/etc/alternatives/clang-tidy -> /usr/bin/clang-tidy-18)
        clang_tidy_exe = util.find_executable_from_system('clang-tidy', follow_symlinks=True)
        if clang_tidy_exe:
            self.clang_tidy_path = clang_tidy_exe
            if self.print: console(f'Found clang-tidy in PATH and resolved as: {clang_tidy_exe}', color=Color.GREEN)
            return

        self.clang_tidy_path = None
        console('clang-tidy not found! Static analysis will be disabled.', color=Color.YELLOW)
        console('install clang-tidy and add to PATH or define env CLANG_TIDY=<path>', color=Color.YELLOW)


    def add_sanitizer_option(self, option):
        if self.sanitize: self.sanitize += ',' + option
        else:             self.sanitize = option


    def add_coverage_option(self, option='default'):
        if self.coverage: self.coverage += ',' + option
        else:             self.coverage = option


    def append_env_path(self, paths, env):
        path = os.getenv(env)
        if path: paths.append(path)


    def raspi_bin(self):
        if not self.raspi_compilers: self.init_raspi_path()
        return self.raspi_compilers


    def raspi_sysroot(self):
        if not self.raspi_compilers: self.init_raspi_path()
        return self.raspi_system


    def raspi_includes(self):
        if not self.raspi_compilers: self.init_raspi_path()
        return self.raspi_include_paths


    def init_raspi_path(self):
        paths = []
        self.append_env_path(paths, 'RASPI_HOME')
        self.append_env_path(paths, 'RASPBERRY_HOME')
        if System.windows: paths += ['/SysGCC/raspberry']
        elif System.linux: paths += ['/usr/bin/raspberry', '/usr/local/bin/raspberry', '/opt/raspberry']
        compiler = ''
        if System.windows: compiler = 'bin/arm-linux-gnueabihf-gcc.exe'
        elif System.linux: compiler = 'arm-bcm2708/arm-linux-gnueabihf/bin/arm-linux-gnueabihf-gcc'
        for raspi_path in paths:
            if os.path.exists(f'{raspi_path}/{compiler}'):
                if not System.windows:
                    raspi_path = f'{raspi_path}/arm-bcm2708/arm-linux-gnueabihf/'
                self.raspi_compilers = f'{raspi_path}/bin/'
                self.raspi_system    = f'{raspi_path}/arm-linux-gnueabihf/sysroot'
                self.raspi_include_paths = [f'{raspi_path}/arm-linux-gnueabihf/lib/include']
                if self.print: console(f'Found RASPI TOOLS: {self.raspi_compilers}\n    sysroot: {self.raspi_system}')
                return
        raise EnvironmentError(f'''No Raspberry PI toolchain compilers detected!
Default search paths: {paths}
Define env RASPI_HOME with path to Raspberry tools.''')


    def set_android_toolchain(self, toolchain_file):
        """
        Sets the toolchain file for Android NDK.
        This should be abspath such as `/opt/android-sdk/ndk/25.2.9519653/build/cmake/android.toolchain.cmake`
        """
        self.android.set_toolchain_path(toolchain_file)


    def set_yocto_toolchain(self, toolchain_dir=None, toolchain_file=None):
        """
        For i.MX8M Plus, Xilinx, Oclea and other Yocto Linux based platforms.
        Sets the toolchain dir where these subdirs exist:
            aarch64-poky-linux/
            x86_64-pokysdk-linux/
        And optionally also sets the CMake toolchain file via `toolchain_file`.
        The `toolchain_file` is only used if `toolchain_dir` chosen as valid.
        """
        self.yocto_linux.init_toolchain(toolchain_dir, toolchain_file)


    def set_oclea_toolchain(self, toolchain_dir=None, toolchain_file=None):
        """
        Sets the toolchain dir where these subdirs exist:
            aarch64-oclea-linux/
            x86_64-ocleasdk-linux/
        And optionally also sets the CMake toolchain file via `toolchain_file`.
        The `toolchain_file` is only used if `toolchain_dir` chosen as valid.
        """
        self.yocto_linux.init_toolchain(toolchain_dir, toolchain_file)


    def set_imx8mp_toolchain(self, toolchain_dir=None, toolchain_file=None):
        """
        Sets the toolchain dir where these subdirs exist:
            aarch64-imx8mp-linux/
            x86_64-imx8mp-sdk-linux/
        And optionally also sets the CMake toolchain file via `toolchain_file`.
        The `toolchain_file` is only used if `toolchain_dir` chosen as valid.
        """
        self.yocto_linux.init_toolchain(toolchain_dir, toolchain_file)


    def set_xilinx_toolchain(self, toolchain_dir=None, toolchain_file=None):
        """
        Sets the toolchain dir where these subdirs exist:
            aarch64-xilinx-linux/
            x86_64-petalinux-linux/
        And optionally also sets the CMake toolchain file via `toolchain_file`.
        The `toolchain_file` is only used if `toolchain_dir` chosen as valid.
        """
        self.yocto_linux.init_toolchain(toolchain_dir, toolchain_file)


    def set_mips_toolchain(self, arch, toolchain_dir=None, toolchain_file=None):
        """
        Sets the toolchain dir for MIPS platform where at least the `bin` dir should exist
        And optionally also sets the CMake toolchain file via `toolchain_file`
        """
        self.mips.init_toolchain(arch, toolchain_dir, toolchain_file)


    def find_default_fortran_compiler(self):
        paths = []
        if System.linux:
            paths += [util.find_executable_from_system('gfortran')]

        for fortran_path in paths:
            if fortran_path and os.path.exists(fortran_path):
                if self.verbose: console(f'Found Fortran: {fortran_path}')
                return fortran_path
        return None


    def get_visualstudio_path(self):
        if self._visualstudio_path:
            return self._visualstudio_path
        if not System.windows:
            raise EnvironmentError('VisualStudio tools support not available on this platform!')

        vswhere_exe = "C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\vswhere.exe"
        vspath = execute_piped(f'"{vswhere_exe}" -latest -nologo -property installationPath')
        if vspath and os.path.exists(vspath):
            self._visualstudio_path = vspath
            if self.verbose: console(f'Detected VisualStudio: {vspath}')
            return vspath

        paths = []
        vs_variants = [ 'Enterprise', 'Professional', 'Community'  ]
        for version in [ '18', '2022' ]: # new 64-bit VS
            for variant in vs_variants:
                paths.append(f'C:\\Program Files\\Microsoft Visual Studio\\{version}\\{variant}')
        for version in [ '2019', '2017' ]:
            for variant in vs_variants:
                paths.append(f'C:\\Program Files (x86)\\Microsoft Visual Studio\\{version}\\{variant}')

        for path in paths:
            if path and os.path.exists(path):
                self._visualstudio_path = path
                if self.verbose: console(f'Detected VisualStudio: {path}')
                return path

        return self._visualstudio_path


    def is_target_arch_x64(self): return self.arch == 'x64'
    def is_target_arch_x86(self): return self.arch == 'x86'
    def is_target_arch_arm64(self): return self.arch == 'arm64'
    def is_target_arch_armv7(self): return self.arch == 'arm'


    def get_gcc_linux_march(self):
        if self.is_target_arch_arm64(): return 'native' if System.aarch64 else 'armv8-a'
        if self.is_target_arch_x64(): return 'native' if System.x86_64 else 'x86-64'
        if self.is_target_arch_x86(): return 'native' if System.x86 else 'pentium4'
        raise RuntimeError(f'Unsupported arch: {self.arch}')


    def get_visualstudio_cmake_arch(self):
        if self.is_target_arch_x64(): return 'x64'
        if self.is_target_arch_x86(): return 'Win32'
        if self.arch == 'arm':   return 'ARM'
        if self.arch == 'arm64': return 'ARM64'
        raise RuntimeError(f'Unsupported arch: {self.arch}')


    def get_visualstudio_cmake_id(self):
        if self._visualstudio_cmake_id:
            return self._visualstudio_cmake_id

        path = self.get_visualstudio_path()
        if '\\2022\\' in path: self._visualstudio_cmake_id = 'Visual Studio 17 2022'
        elif '\\18\\' in path: self._visualstudio_cmake_id = 'Visual Studio 18 2026'
        elif '\\2019\\' in path: self._visualstudio_cmake_id = 'Visual Studio 16 2019'
        else:                  self._visualstudio_cmake_id = 'Visual Studio 15 2017'

        if self.verbose: console(f'Detected CMake Generator: -G"{self._visualstudio_cmake_id}" -A {self.get_visualstudio_cmake_arch()}')
        return self._visualstudio_cmake_id


    def get_msbuild_path(self):
        if self._msbuild_path:
            return self._msbuild_path

        paths = [ util.find_executable_from_system('msbuild') ]
        if System.windows:
            vswhere = '"C:\\Program Files (x86)\\Microsoft Visual Studio\\Installer\\vswhere.exe" -latest -nologo -property installationPath'
            paths.append(f"{execute_piped(vswhere)}\\MSBuild\\Current\\Bin\\MSBuild.exe")
            paths.append(f"{execute_piped(vswhere)}\\MSBuild\\15.0\\Bin\\amd64\\MSBuild.exe")

            vs_variants = [ 'Enterprise', 'Professional', 'Community' ]
            for variant in vs_variants:
                paths.append(f'C:\\Program Files (x86)\\Microsoft Visual Studio\\2019\\{variant}\\MSBuild\\Current\\Bin\\MSBuild.exe')
                paths.append(f'C:\\Program Files (x86)\\Microsoft Visual Studio\\2017\\{variant}\\MSBuild\\15.0\\Bin\\amd64\\MSBuild.exe')

        for path in paths:
            if path and os.path.exists(path):
                self._msbuild_path = path
                if self.verbose: console(f'Detected MSBuild: {path}')
                return path
        raise EnvironmentError('Failed to find MSBuild from system PATH. You can easily configure msbuild by running `mama install-msbuild`.')


    ## MSVC tools at, for example: "{VisualStudioPath}\VC\Tools\MSVC\14.16.27023"
    def get_msvc_tools_path(self):
        if self._msvctools_path:
            return self._msvctools_path
        if not System.windows:
            raise EnvironmentError('MSVC tools not available on this platform!')

        tools_root = f"{self.get_visualstudio_path()}\\VC\\Tools\\MSVC"
        tools = os.listdir(tools_root)
        if not tools:
            raise EnvironmentError('Could not detect MSVC Tools')

        tools_path = os.path.join(tools_root, tools[0])
        #tools_path = forward_slashes(tools_path)
        self._msvctools_path = tools_path
        if self.verbose: console(f'Detected MSVC Tools: {tools_path}')
        return tools_path


    def get_msvc_bin64(self):
        return f'{self.get_msvc_tools_path()}/bin/Hostx64/x64/'


    def get_msvc_link64(self):
        return f'{self.get_msvc_bin64()}link.exe'


    def get_msvc_cl64(self):
        return f'{self.get_msvc_bin64()}cl.exe'


    def get_msvc_lib64(self):
        return f'{self.get_msvc_tools_path()}\\lib\\x64'


    def install_clang(self, clang_major):
        if type(clang_major) != int: clang_major = int(clang_major) # convert to int
        if System.windows: raise OSError('Install Visual Studio 2026 with Clang support')
        if System.macos:   raise OSError('Install Xcode to get Clang on macOS')
        id, major, minor = self.get_distro_info()
        if id != "ubuntu": raise OSError(f'install-clang-{clang_major} only supports ubuntu')
        console(f'Installing clang-{clang_major} and libc++-{clang_major}-dev from apt repositories', color=Color.MAGENTA)
        execute('sudo apt-get update')
        execute(f'sudo apt-get install clang-{clang_major} clang-tidy-{clang_major} '+\
                f'libc++-{clang_major}-dev libc++abi-{clang_major}-dev -y')
        # configure current clang version as default clang via update-alternatives
        # this way mama and cmake tools can find it without additional configuration
        console(f'Configuring clang-{clang_major} as default clang via update-alternatives', color=Color.MAGENTA)
        execute(f'sudo update-alternatives --install /usr/bin/clang   clang   /usr/bin/clang-{clang_major}   100')
        execute(f'sudo update-alternatives --install /usr/bin/clang++ clang++ /usr/bin/clang++-{clang_major} 100')
        execute(f'sudo update-alternatives --install /usr/bin/clang-tidy clang-tidy /usr/bin/clang-tidy-{clang_major} 100')
        execute(f'sudo update-alternatives --set clang   /usr/bin/clang-{clang_major}')
        execute(f'sudo update-alternatives --set clang++ /usr/bin/clang++-{clang_major}')
        execute(f'sudo update-alternatives --set clang-tidy /usr/bin/clang-tidy-{clang_major}')

    
    def install_gcc(self, gcc_major):
        if type(gcc_major) != int: gcc_major = int(gcc_major) # convert to int
        if System.windows: raise OSError('Install MinGW to get GCC on Windows')
        if System.macos:   raise OSError('install-gcc not implemented for macOS')
        id, major, minor = self.get_distro_info()
        if id != "ubuntu": raise OSError(f'install-gcc-{gcc_major} only supports ubuntu')
        console(f'Installing gcc-{gcc_major} and g++-{gcc_major} from apt repositories', color=Color.MAGENTA)
        execute('sudo apt-get update')
        execute(f'sudo apt-get install gcc-{gcc_major} g++-{gcc_major} -y')
        # configure current gcc version as default gcc via update-alternatives
        # this way mama and cmake tools can find it without additional configuration
        console(f'Configuring gcc-{gcc_major} as default gcc via update-alternatives', color=Color.MAGENTA)
        execute(f'sudo update-alternatives --install /usr/bin/gcc gcc /usr/bin/gcc-{gcc_major} 100')
        execute(f'sudo update-alternatives --install /usr/bin/g++ g++ /usr/bin/g++-{gcc_major} 100')
        execute(f'sudo update-alternatives --set gcc /usr/bin/gcc-{gcc_major}')
        execute(f'sudo update-alternatives --set g++ /usr/bin/g++-{gcc_major}')


    def install_msbuild(self):
        if System.windows: raise OSError('Install Visual Studio 2019 to get MSBuild on Windows')
        if System.macos:   raise OSError('install_msbuild not implemented for macOS')

        id, _, _ = self.get_distro_info()
        if id != "ubuntu": raise OSError('install_msbuild only supports ubuntu')
        codename = distro.info()['codename']

        execute('curl https://packages.microsoft.com/keys/microsoft.asc | gpg --dearmor > /tmp/microsoft.gpg')
        execute('sudo mv /tmp/microsoft.gpg /etc/apt/trusted.gpg.d/microsoft.gpg')
        execute(f"sudo sh -c 'echo \"deb [arch=amd64] https://packages.microsoft.com/repos/microsoft-ubuntu-{codename}-prod {codename} main\" > /etc/apt/sources.list.d/dotnetdev.list'")
        execute('sudo apt-get install apt-transport-https')
        execute('sudo apt-get update')
        execute('sudo apt-get install dotnet-sdk-2.1')


    def install_ndk(self):
        if System.macos:
            raise OSError('install_ndk not implemented for macOS')
        elif System.windows:
            ndk_version = '25.2.9519653'
            ndk_url = 'https://dl.google.com/android/repository/android-ndk-r25c-windows.zip'
            ndk_dest = f'{os.getenv("LOCALAPPDATA")}\\Android\\Sdk\\ndk'
        elif System.linux:
            ndk_version = '25.2.9519653'
            ndk_url = 'https://dl.google.com/android/repository/android-ndk-r25c-linux.zip'
            ndk_dest = f'/opt/android-sdk/ndk'

        console(f'Downloading NDK {ndk_version}')
        ndk_zip = util.download_file(ndk_url, tempfile.gettempdir())

        if System.windows:
            os.makedirs(ndk_dest, exist_ok=True)
        else:
            execute(f'sudo mkdir -p {ndk_dest} && sudo chown -R $USER {ndk_dest}')

        console(f'Extracting NDK to {ndk_dest}/{ndk_version}')
        util.unzip(ndk_zip, ndk_dest)

        final_dest = f'{ndk_dest}/{ndk_version}'
        if os.path.exists(final_dest):
            shutil.rmtree(final_dest)

        shutil.move(f'{ndk_dest}/android-ndk-r25c', final_dest)
        if os.path.exists(f'{final_dest}/build'):
            console(f'NDK installed successfully to {final_dest}')
        else:
            raise RuntimeError(f'Failed to install NDK to {final_dest}')

        console(f'Adding ANDROID_NDK_HOME={final_dest} to ~/.bashrc, run source ~/.bashrc or restart terminal to populate your env.')
        execute(f'echo "export ANDROID_NDK_HOME={final_dest}" >> ~/.bashrc')


    def run_convenient_installs(self):
        for tool in self.convenient_install:
            if 'clang-' in tool: self.install_clang(tool[6:])
            elif 'gcc-' in tool: self.install_gcc(tool[4:])
            elif 'msbuild' in tool: self.install_msbuild()
            elif 'ndk'     in tool: self.install_ndk()


    def libname(self, library):
        if self.msvc: return f'{library}.lib'
        else:         return f'lib{library}.a'


    def libext(self):
        return 'lib' if self.msvc else 'a'


    def has_target(self) -> bool:
        """ A target was specified from cmdline, eg 'all' or 'mypackage' """
        return self.target is not None and len(self.target) > 0

    def no_target(self) -> bool:
        """ No target specified from cmdline """
        return self.target is None or len(self.target) == 0

    def targets_all(self) -> bool:
        """ Target specified from cmdline was 'all' """
        return self.target == 'all'


    def target_matches(self, target_name: str) -> bool:
        """ True if target_name matches the target specified from cmdline """
        return self.targets_all() \
            or (self.target and self.target.lower() == target_name.lower())


    def no_specific_target(self) -> bool:
        """ True if no target or 'all' was specified from cmdline """
        return self.no_target() or self.targets_all()




from __future__ import annotations
from typing import TYPE_CHECKING

import os, shutil, stat, string
from .dep_source import DepSource
from ..utils.system import Color, System, console, error
from ..utils.sub_process import SubProcess, execute, execute_piped, execute_piped_echo
from ..util import is_dir_empty, save_file_if_contents_changed, read_lines_from, path_join


if TYPE_CHECKING:
    from ..build_target import BuildTarget
    from ..build_config import BuildConfig
    from ..build_dependency import BuildDependency

class Git(DepSource):
    """
    For BuildDependency whose source is from a Git repository
    """
    def __init__(self, name:str, url:str, branch:str, tag:str, mamafile:str, shallow:bool, args:list):
        super(Git, self).__init__(name)
        if not url: raise RuntimeError("Git url must not be empty!")
        self.is_git = True
        self.url = url
        self.branch = branch
        self.tag = tag
        self.mamafile = mamafile
        self.shallow = shallow
        self.args = args

        self.from_source = False  # if True, this must be built from source, not from artifactory
        self.commit_hash = None  # the git commit hash of this DepSource

        self.missing_status = False
        self.url_changed = False
        self.tag_changed = False
        self.branch_changed = False
        self.commit_changed = False


    def __repr__(self): return self.__str__()
    def __str__(self):
        s = f'DepSource Git {self.name} {self.url}'
        tag = self.branch_or_tag()
        if tag: s += ' ' + tag
        if self.mamafile: s += ' ' + self.mamafile
        return s

    @staticmethod
    def from_papa_string(s: str) -> "Git":
        p = s.split(',')
        name, url, branch, tag, mamafile = p[0:5]
        args = p[5:]
        shallow = True # shallow is the default
        return Git(name, url, branch, tag, mamafile, shallow, args)


    def get_papa_string(self):
        fields = DepSource.papa_join(
            self.name, self.url, self.branch, self.tag, self.mamafile, self.args)
        return 'git ' + fields


    def run_git(self, dep: BuildDependency, git_command, throw=True):
        cmd = f"cd {dep.src_dir} && git {git_command}"
        if dep.config.verbose:
            console(f'  {dep.name: <16} git {git_command}', color=Color.YELLOW)
        return execute(cmd, throw=throw)


    def _has_local_modifications(self, dep: BuildDependency) -> bool:
        """Check if the working tree has uncommitted modifications to tracked files"""
        return self.run_git(dep, "diff --quiet HEAD", throw=False) != 0


    def get_commit_hash(self, dep: BuildDependency, use_cache=True):
        if not self.commit_hash or not use_cache:
            self.commit_hash = self.init_commit_hash(dep, use_cache=use_cache, fetch_remote=True)
        return self.commit_hash

    @staticmethod
    def get_current_repository_commit(dep: BuildDependency):
        """ Assuming {src_dir}/.git exists, this will get the repository commit short hash """
        result = execute_piped(['git', 'show', '--format=%h', '-s'], cwd=dep.src_dir)
        if dep.config.verbose:
            console(f'  {dep.name: <16} git show --format=%h -s:   {result}')
        return result

    @staticmethod
    def is_hex_string(s: str) -> bool:
        return len(s) > 0 and all(c in string.hexdigits for c in s)

    def init_commit_hash(self, dep: BuildDependency, use_cache: bool, fetch_remote: bool):
        """
        Gets the latest commit hash, based on git source tag and branch options.
        """
        if not dep.dep_source.is_git:
            return None

        # update is not specified? then we can try to skip the check
        if use_cache and not dep.config.update and os.path.exists(self.git_status_file(dep)):
            status = self.read_stored_status(dep)
            result = status[3].split(' ')[0]
            if dep.config.verbose:
                console(f'    {self.name}  using stored commit hash: {result}')
            return result

        # is the tag actually a commit hash?
        if Git.is_hex_string(self.tag):
            if dep.config.verbose:
                console(f'    {self.name}  using tag as the commit hash: {self.tag}')
            return self.tag

        # is this a git repository? we can get the current commit from that
        if os.path.exists(f'{dep.src_dir}/.git'):
            result = Git.get_current_repository_commit(dep)
            if not result:
                error(f'    {self.name}  invalid git repository at {dep.src_dir}')
            return result

        # can we fetch the latest commit from remote instead?
        if fetch_remote:
            arguments = 'HEAD'
            try:
                if self.branch: arguments = self.branch
                elif self.tag:  arguments = self.tag
                result = execute_piped(f'git ls-remote {self.url} {arguments}', timeout=5)
                if result: result = result.split(' ')[0][0:7]
                if dep.config.verbose:
                    console(f'    {self.name}  git ls-remote {self.url} {arguments}: {result}', color=Color.YELLOW)
                return result
            except Exception as e:
                if dep.config.verbose:
                    error(f'    {self.name}  git ls-remote {self.url} {arguments} failed: {e}')
                return None


    def _is_detached_head(self, dep: BuildDependency) -> bool:
        """Check if the repository is in detached HEAD state"""
        result = execute_piped(['git', 'symbolic-ref', '-q', 'HEAD'], cwd=dep.src_dir, throw=False)
        return not result


    def fetch_origin(self, dep: BuildDependency):
        branch = self.branch_or_tag()
        if Git.is_hex_string(branch):
            return # no need to fetch if we're pinned to a specific commit hash
        if self.tag:
            self.run_git(dep, f"fetch origin tag {branch} -q")
        else:
            # only safe to pull when staying on the same branch and not in detached HEAD
            can_pull = not (self.tag_changed or self.branch_changed or self._is_detached_head(dep))
            origin_br = f'origin {branch}' if branch else 'origin'
            result = -1
            if can_pull:
                result = self.run_git(dep, f"pull {origin_br} -q", throw=False)
            if result != 0:
                self.run_git(dep, f"fetch {origin_br} -q")


    def git_status_file(self, dep: BuildDependency):
        return path_join(dep.build_dir, 'git_status')

    @staticmethod
    def format_git_status(url: str, tag: str, branch: str, commit: str):
        return f"{url}\n{tag}\n{branch}\n{commit}\n"

    def save_status(self, dep: BuildDependency):
        commit = self.get_commit_hash(dep)
        status = self.format_git_status(self.url, self.tag, self.branch, commit)
        if save_file_if_contents_changed(self.git_status_file(dep), status):
            if dep.config.verbose:
                console(f'    {self.name}  write git status commit={commit}')


    def read_stored_status(self, dep: BuildDependency):
        lines = read_lines_from(self.git_status_file(dep))
        if not lines: return None
        url = lines[0].rstrip()
        tag = lines[1].rstrip()
        branch = lines[2].rstrip()
        commit = lines[3].rstrip()
        return (url, tag, branch, commit)


    def reset_status(self, dep: BuildDependency):
        """ Clears the status file """
        self.missing_status = True
        status_file = self.git_status_file(dep)
        if os.path.exists(status_file):
            os.remove(status_file)


    def check_status(self, dep: BuildDependency):
        status = self.read_stored_status(dep)
        if not status:
            self.missing_status = True
            if not self.url: return False
            #console(f'check_status {self.url}: NO STATUS AT {dep.build_dir}/git_status')
            self.url_changed = True
            self.tag_changed = True
            self.branch_changed = True
            self.commit_changed = True
            return True
        # update basic flags before fetching origin, so we can detect tag/branch pin changes
        self.url_changed = self.url != status[0]
        self.tag_changed = self.tag != status[1]
        self.branch_changed = self.branch != status[2]
        # fetch origin and then check the commit hash to detect upstream changes
        self.fetch_origin(dep)
        self.commit_changed = self.get_commit_hash(dep, use_cache=False) != status[3]
        #console(f'check_status {self.url} {self.branch_or_tag()}: urlc={self.url_changed} tagc={self.tag_changed} brnc={self.branch_changed} cmtc={self.commit_changed}')
        return self.url_changed or self.tag_changed or self.branch_changed or self.commit_changed


    def branch_or_tag(self):
        if self.branch: return self.branch
        if self.tag: return self.tag
        return ''


    def checkout_current_branch_or_tag(self, dep: BuildDependency, is_commit_pin=False):
        branch = self.branch_or_tag()
        if branch:
            if self.tag_changed or self.branch_changed:
                self.run_git(dep, "rebase --abort", throw=False)
                self.run_git(dep, "reset --hard")
            if is_commit_pin:
                self.run_git(dep, f"fetch --depth 1 origin {branch}")
                self.run_git(dep, f"checkout {branch}")
            elif self.branch:
                # hacky fix
                self.run_git(dep, f"fetch origin +refs/heads/{branch}:refs/remotes/origin/{branch}", throw=False)
                self.run_git(dep, f"checkout -B {branch} origin/{branch}")
            else: # tag
                if self.tag_changed:
                    self.run_git(dep, f"fetch origin tag {branch}")
                self.run_git(dep, f"checkout {branch}")


    def reclone_wipe(self, dep: BuildDependency):
        if dep.config.print:
            console(f'  - Target {dep.name: <16} RECLONE WIPE')
        if os.path.exists(dep.dep_dir):
            if System.windows: # chmod everything to user so we can delete:
                for root, dirs, files in os.walk(dep.dep_dir):
                    for d in dirs:  os.chmod(os.path.join(root, d), stat.S_IWUSR)
                    for f in files: os.chmod(os.path.join(root, f), stat.S_IWUSR)
            shutil.rmtree(dep.dep_dir)


    def clone_with_filtered_progress(self, dep: BuildDependency, clone_args: str, clone_to_dir: str):
        output = ''
        current_percent = -1
        def print_output(p:SubProcess, line:str):
            nonlocal output, current_percent
            if 'remote: Counting objects:' in line or \
                'remote: Compressing objects:' in line or \
                'Receiving objects:' in line or \
                'Resolving deltas:' in line or \
                'Updating files:' in line:
                if dep.config.print:
                    parts = line.split('%')[0].split(':')
                    if not parts:
                        console(line)
                        return
                    percent_string = parts[len(parts)-1].strip()
                    percent = int(percent_string) if percent_string else 0
                    if current_percent != percent:
                        current_percent = percent
                        status = 'status             '
                        if 'remote: Counting objects:' in line:      status = 'counting objects   '
                        elif 'remote: Compressing objects:' in line: status = 'compressing objects'
                        elif 'Receiving objects:' in line:           status = 'receiving objects  '
                        elif 'Resolving deltas:' in line:            status = 'resolving deltas   '
                        elif 'Updating files:' in line:              status = 'updating files     '
                        console(f'\r  - Target {dep.name: <16} CLONE {status} {current_percent:3}%', end='')
            elif 'Cloning into ' in line:
                return
            elif 'Are you sure you want to continue connecting' in line:
                # TODO: maybe auto-add the key before running clone?
                # if [ ! -n "$(grep "^bitbucket.org " ~/.ssh/known_hosts)" ]; then ssh-keyscan bitbucket.org >> ~/.ssh/known_hosts 2>/dev/null; fi
                console(line)
                p.write('yes\n') # get us unstuck
            elif line:
                output += line
                output += '\n'
                if dep.config.verbose:
                    console(line)

        # run the command, working dir not needed since it should be a full path in the clone_args
        cmd = f'git clone {clone_args} {clone_to_dir}'
        if dep.config.verbose:
            console(f'  {dep.name: <16} {cmd}')
        result = SubProcess.run(cmd, io_func=print_output)

        # handle the result:
        if dep.config.print:
            if result == 0:
                console(f'\r  - Target {dep.name: <16} CLONE SUCCESS                  ', color=Color.BLUE)
                if dep.config.verbose and output:
                    console(output, end='')
            else:
                console(f'\r  - Target {dep.name: <16} CLONE FAILED ({result})              ', color=Color.RED)
                if output:
                    console(output, end='')
                raise RuntimeError(f'Target {self.name} clone failed: {cmd}')


    def clone_or_pull(self, dep: BuildDependency, wiped=False):
        # by default we create a shallow clone, unless unshallow is specified in config or this dep
        unshallow = dep.config.unshallow or (not self.shallow)
        if is_dir_empty(dep.src_dir):
            if not wiped and dep.config.print:
                console(f"  - Target {dep.name: <16} CLONE because src is missing", color=Color.BLUE)
            br_or_tag = self.branch_or_tag()
            is_commit_pin = Git.is_hex_string(br_or_tag)
            checkout_branch = '' if is_commit_pin or len(br_or_tag) == 0 else f' --branch {br_or_tag}'
            depth = '' if unshallow else '--depth 1'
            clone_args = f"--recurse-submodules {depth} {checkout_branch} {self.url}"
            self.clone_with_filtered_progress(dep, clone_args, dep.src_dir)
            self.checkout_current_branch_or_tag(dep, is_commit_pin=is_commit_pin)
        else:
            if dep.config.print:
                console(f"  - Pulling {dep.name: <16}  SCM change detected", color=Color.BLUE)
            # check for local modifications before potentially destructive operations
            if self._has_local_modifications(dep):
                name = dep.name
                error(f"  Target {name} has local modifications that would be overwritten by update.\n"
                      f"  To discard local changes and re-fetch, run: `mama reclone {name}`")
                self.run_git(dep, "status --porcelain") # show the user what files are modified
                raise RuntimeError(f"Target {name} has local modifications. Use 'mama reclone {name}' to discard changes.")
            if unshallow:
                self.unshallow(dep)
            is_commit_pin = Git.is_hex_string(self.branch_or_tag())
            self.checkout_current_branch_or_tag(dep, is_commit_pin=is_commit_pin)
            self.run_git(dep, 'submodule update --init --recursive')
            if not self.tag: # pull if not a tag
                if self.branch:
                    self.run_git(dep, f"fetch origin {self.branch} -q", throw=False)
                    self.run_git(dep, f"reset --hard origin/{self.branch} -q")
                else:
                    self.run_git(dep, "fetch -q", throw=False)
                    self.run_git(dep, "reset --hard @{upstream} -q") # https://git-scm.com/docs/gitrevisions#Documentation/gitrevisions.txt-branchnameupstreamegmasterupstreamu


    def unshallow(self, dep: BuildDependency):
        # Detecting shallowness can be quite tricky, since the repository can be in multiple different states
        # Detecting the easy cases first:
        #   git rev-parse --is-shallow-repository --> .git/shallow exists
        is_shallow = os.path.exists(f'{dep.src_dir}/.git/shallow')
        if not is_shallow:
            _, output = execute_piped_echo(dep.src_dir, 'git config remote.origin.fetch', echo=False)
            if dep.config.verbose:
                console(f'  {dep.name: <16} remote.origin.fetch: {output.strip()}', color=Color.YELLOW)
            if not output or not output.startswith('+refs/heads/*'):
                is_shallow = True # likely a shallow clone

        if is_shallow:
            if dep.config.print:
                console(f'  - Unshallowing {dep.name}', color=Color.YELLOW)
            self.run_git(dep, 'config remote.origin.fetch "+refs/heads/*:refs/remotes/origin/*"')
            self.run_git(dep, 'remote update')
            # this last step is allowed to fail, just in case it was
            # semi-shallow (history was complete, but remote refs were shallow)
            self.run_git(dep, 'fetch --unshallow', throw=False)


    def dependency_checkout(self, dep: BuildDependency):
        """
        Do a git repository checkout. Can be an expensive operation.
        If an existing artifactory package exists, then this step is skipped
        """
        if not dep.source_dir_exists():  # we MUST pull here
            self.clone_or_pull(dep)
            return True

        is_target = dep.is_current_target()
        config = dep.config
        changed = False

        if config.update and is_target:
            changed = self.check_status(dep)

        wiped = False
        should_wipe = self.url_changed and not self.missing_status
        if should_wipe or (is_target and config.reclone):
            self.reclone_wipe(dep)
            wiped = True
        elif dep.config.unshallow and is_target: # unshallow was specified, we should at least pull
            pass # fallthrough to clone_or_pull
        else:
            # don't pull if no changes to git status
            # or if we're current target of a non-update build
            # mama update target=ReCpp  -- this should git pull
            # mama build target=ReCpp   -- should NOT pull
            non_update_target = is_target and not config.update
            if non_update_target or not changed:
                if config.verbose:
                    console(f'    {self.name} git no changes detected and update not specified', color=Color.YELLOW)
                return False

        self.clone_or_pull(dep, wiped)
        return True




from __future__ import annotations
from typing import TYPE_CHECKING
import os
from .system import console
from .gdb import filter_gdb_arg, run_gdb
from .run import run_in_working_dir

if TYPE_CHECKING:
    from ..build_target import BuildTarget

def run_gtest(target: BuildTarget, executable: str, args='', src_dir=False, gdb=False):
    args, gdb = filter_gdb_arg(args, gdb)
    ## gtest flags:
    # https://github.com/google/googletest/blob/main/googletest/src/gtest.cc#L238
    params = f' --gtest_output="xml:{target.source_dir("test/report.xml")}"'
    if args:
        for arg in args.split(' '):
            if arg.startswith('--gtest_'):
                params += f' {arg}'
            else:
                params += f' --gtest_filter="*{arg}*"'

    if gdb:
        run_gdb(target, f'{executable} {params}', src_dir=src_dir)
    else:
        dirname = os.path.dirname(executable)
        exename = os.path.basename(executable)
        dir = target.source_dir(dirname) if src_dir else target.build_dir(dirname)
        run_in_working_dir(target, dir, f'{dir}/{exename} {params}')



#!/usr/bin/python3.10
import sys, os

from .types.local_source import LocalSource
from .utils.system import Color, console
from .utils.sub_process import execute
from .util import glob_with_extensions, glob_folders_with_name_match
from .build_config import BuildConfig
from .build_target import BuildTarget
from .build_dependency import BuildDependency
from .dependency_chain import load_dependency_chain, execute_task_chain, find_dependency, get_flat_deps, get_deps_that_depend_on_target
from .init_project import mama_init_project
from ._version import __version__

def print_title():
    console(f'========= Mama Build Tool ==========')

def print_usage():
    console('mama [actions...] [args...]')
    console('  actions:')
    console('    init       - create initial mamafile.py and CMakeLists.txt')
    console('    list       - list all mama dependencies on this project')
    console('    build      - configure and build main project or specific target, this can clone, but does not pull')
    console('    update     - update and build target dependencies after calling git pull')
    console('    deploy     - runs PAPA deploy stage by gathering all libraries and assets')
    console('    serve      - Equivalent of `update rebuild deploy upload`')
    console('    clean      - clean main project or specific target')
    console('    rebuild    - clean, update and build main project or specific target')
    console('    reclone    - wipe specific target dependency and clone it again')
    console('    wipe       - alias of reclone')
    console('    dirty      - mark a target for rebuild even if it was up to date')
    console('    upload     - uploads target package to artifactory server')
    console('    if_needed  - only uploads if package does not exist on server')
    console('    art        - always fetch pkgs from artifactory, failure will throw an error')
    console('    noart      - temporarily ignore artifactory pkgs fetch')
    console('    test       - run tests for main project or specific target')
    console('    start=arg  - start a specific tool via mamafile.start(args)')
    console('    add        - add new dependency')
    console('    new        - create new mama build file')
    console('    open=<tgt> - open a project file')
    console('    version    - shows this package version and exits')
    console('    help       - shows this help list')
    console('  install utils:')
    console('    install-clang-<ver> - configures and installs clang-<ver> for ubuntu, ex: install-clang-18')
    console('    install-gcc-<ver>   - configures and installs gcc-<ver> for ubuntu, ex: install-gcc-13')
    console('    install-msbuild - configures and installs MSBuild for linux')
    console('    install-ndk     - configures and installs Android NDK for linux or windows')
    console('  args:')
    console('    windows    - build for windows (alias for msvc)')
    console('    msvc       - build for windows using MSVC')
    console('    linux      - build for linux')
    console('    imx8mp     - build for nxp imx8mp yocto')
    console('    xilinx     - build for amd xilinx yocto')
    console('    oclea      - build for oclea/ambarella yocto')
    console('    raspi      - build for raspi')
    console('    mips       - build for mips architecture')
    console('    macos      - build for macos')
    console('    ios        - build for ios')
    console('    android    - build for android')
    console('    android-N  - build for android targeting specific API level, ex: android-26')
    console('    clang      - prefer clang for linux (default on linux/macos/ios/android)')
    console('    gcc        - prefer gcc for linux')
    console('    fortran    - enable automatic fortran detection (or configure this in mamafile)')
    console('    release    - (default) CMake configuration RelWithDebInfo')
    console('    debug      - CMake configuration Debug')
    console('    arch=x86   - Override cross-compiling architecture: (x86, x64, arm, arm64)')
    console('    x86        - Shorthand for arch=x86, all shorthands: x86 x64 arm arm64')
    console('    jobs=N     - Max number of parallel compilations. (default=system.core.count)')
    console('    target=P   - Name of the target')
    console('    all        - Short for target=all')
    console('    with_tests - Forces CMake option -DENABLE_TESTS=ON and -DBUILD_TESTS=ON')
    console('    test_until_failure - Runs tests in a loop until they fail, defaults to N=1000, useful to catch flaky tests')
    console('    test_until_failure=N - Runs tests in a loop until they fail, with a maximum of N iterations')
    console('    sanitize=  - enables -fsanitize= for gcc/clang builds [address|leak|thread|undefined]')
    console('    asan|lsan|tsan|ubsan - shorthands for sanitize=address|leak|thread|undefined respectively')
    console('    clang-tidy - enables clang-tidy static analysis during build, clang-tidy must be in PATH')
    console('    coverage   - Builds the project with GCC --coverage option')
    console('    coverage-report[=src_root] - Generates coverage report using gcovr')
    console('    silent     - Greatly reduces verbosity')
    console('    verbose    - Greatly increases verbosity for build dependencies and cmake')
    console('  examples:')
    console('    mama init                      Initialize a new project. Tries to create mamafile.py and CMakeLists.txt')
    console('    mama build                     Update and build main project only. This only clones, but does not update!')
    console('    mama build x86 opencv          Cross compile build target opencv to x86 architecture')
    console('    mama build android             Cross compile to arm64 android NDK')
    console('    mama build android-26 arm      Cross compile to armv7 android NDK API level 26')
    console('    mama update                    Update all dependencies by doing git pull and build.')
    console('    mama clean                     Cleans main project only.')
    console('    mama clean x86 opencv          Cleans main project only.')
    console('    mama clean all                 Cleans EVERYTHING in the dependency chain for current arch.')
    console('    mama rebuild                   Cleans, update and build main project only.')
    console('    mama build dep1                Update and build dep1 only.')
    console('    mama update dep1               Update and build the specified target.')
    console('    mama serve android             Update, build and deploy for Android')
    console('    mama wipe dep1                 Wipe target dependency completely and clone again. Does not build!')
    console('    mama upload dep1               Deploys and uploads dependency to Artifactory server.')
    console('    mama test                      Run tests on main project.')
    console('    mama test=arg                  Run tests on main project with an argument.')
    console('    mama test="arg1 arg2"          Run tests on main project with multiple arguments.')
    console('    mama test dep1                 Run tests on target dependency project.')
    console('    mama start=dbtool              Call main project mamafile start() with args [`dbtool`].')
    console('    mama rebuild all tsan lsan     Rebuild all targets with thread and leak sanitizers enabled.')
    console('    mama rebuild all sanitize=leak,undefined Rebuild all targets with leak and undefined sanitizers enabled.')
    console('  environment:')
    console('    setenv("NINJA")                  Path to NINJA build executable')
    console('    setenv("ANDROID_HOME")           Path to Android SDK if auto-detect fails')
    console('    setenv("MAMA_ARTIFACTORY_USER")  Username for Artifactory server')
    console('    setenv("MAMA_ARTIFACTORY_PASS")  Password for Artifactory server')


def open_project(config: BuildConfig, root_dependency: BuildDependency):
    name = config.target if config.has_target() and not config.targets_all() else config.open
    found = root_dependency if name == 'root' else find_dependency(root_dependency, name)
    if not found:
        raise KeyError(f'No project named {name}')
    
    if config.msvc:
        solutions = glob_with_extensions(found.build_dir, ['.sln'])
        if solutions:
            execute(f'start {solutions[0]}', echo=True)
        else:
            console('Could not find any Visual Studio solutions, using VSCode instead.')
            execute(f'code {found.src_dir}', echo=True)

    elif config.macos or config.ios:
        projects = glob_folders_with_name_match(found.build_dir, ['.xcodeproj'])
        if projects:
            execute(f'open {projects[0]}', echo=True)
        else:
            console('Could not find any Xcode projects, using VSCode instead.')
            execute(f'code {found.src_dir}', echo=True)

    elif config.linux:
        console(f'Using VSCode. You can also try opening this folder with CLion: {found.src_dir}')
        execute(f'code {found.src_dir}', echo=True)
        #execute(f'xdg-open', echo=True)

    elif config.android:
        console('Android IDE selection not implemented, using VSCode instead.')
        execute(f'code {found.src_dir}', echo=True)

def set_target_from_unused_args(config: BuildConfig):
    for arg in config.unused_args:
        if config.has_target():
            console(f"ERROR: Deduced Target='{arg}' from unused argument, but target is already set to '{config.target}'")
            exit(-1)
        else:
            config.target = arg

def check_config_target(config: BuildConfig, root: BuildDependency):
    if config.has_target() and not config.targets_all():
        dep = find_dependency(root, config.target)
        if dep is None:
            console(f"ERROR: specified target='{config.target}' not found!")
            exit(-1)


def print_package_exports(dep: BuildDependency):
    target:BuildTarget = dep.target
    if dep.from_artifactory or target.try_automatic_artifactory_fetch():
        console(f'    Target {target.name} fetched from artifactory')
    else:
        target.package()
    target.print_exports(abs_paths=True)


def mama_dirty(root: BuildDependency, dep: BuildDependency):
    """ Marks `dep` as dirty and also marks all projects that depend on `dep` as dirty """
    dirty_chain = get_deps_that_depend_on_target(root, dep)
    if root.config.print:
        used_by = ", ".join([d.name for d in dirty_chain]) if dirty_chain else 'none'
        console(f'    Target {dep.name} used by: {used_by}')
    dep.dirty()
    for d in dirty_chain:
        d.dirty()


def run_coverage_report(target: BuildTarget):
    if target.config.msvc:
        console('Coverage report not supported yet on Windows')
        return
    root = target.source_dir(target.config.coverage_report)
    gcov_exec = ''
    if target.config.gcc and target.config.cc_path:
        # Derive gcov path from gcc path: e.g. /usr/bin/gcc-14 -> /usr/bin/gcov-14
        gcov_path = os.path.realpath(target.config.cc_path).replace('gcc', 'gcov')
        if os.path.exists(gcov_path):
            gcov_exec = f'--gcov-executable "{gcov_path}" '
    verbose = '--verbose ' if target.config.verbose else ''
    cmd = 'gcovr --gcov-ignore-parse-errors negative_hits.warn ' \
        + '--sort uncovered-percent ' \
        + gcov_exec \
        + verbose \
        + f'--root "{root}" "{target.build_dir()}"'
    target.run(cmd, src_dir=True, exit_on_fail=False)


def mamabuild(args, source_dir=os.getcwd()):
    """ Main entry point for MamaBuild. Parses command line arguments and executes the requested actions. 
        - args: list of command line arguments, without the script name. Ex: ['build', 'target=all', 'debug']
        - source_dir: the directory to treat as the main project source
    """
    if sys.version_info < (3, 10):
        console('FATAL ERROR: MamaBuild requires Python 3.10 or higher')
        exit(-1)

    if len(args) == 0 or 'help' in args:
        print_title()
        print_usage()
        exit(-1)
    if 'version' in args:
        console(f'MamaBuild version {__version__}')
        exit(0)

    config = BuildConfig(args)
    if config.print:
        print_title()
        if config.verbose:
            console(f'Build jobs={config.jobs}')

    name = os.path.basename(source_dir)
    local_src = LocalSource(name, source_dir, mamafile=None, always_build=False, args=[])
    workspace = None # figure out the workspace from the root mamafile.py
    root = BuildDependency(None, config, workspace, local_src)

    if config.unused_args:
        set_target_from_unused_args(config)

    # root init
    if config.mama_init and config.no_target():
        mama_init_project(root)
        return

    if config.convenient_install:
        config.run_convenient_installs()
        return

    has_cmake = root.cmakelists_exists()
    if not root.mamafile_exists() and not has_cmake:
        console('FATAL ERROR: mamafile.py not found and CMakeLists.txt not found')
        exit(-1)

    if config.update:
        if config.no_target():
            config.target = 'all'
            if config.print: console(f'Updating all targets')
        else:
            if config.print: console(f'Updating {config.target} target')

    if config.rebuild:
        config.build = True
        config.clean = True

    if config.clean and config.no_target():
        root.clean()

    load_dependency_chain(root)
    check_config_target(config, root)

    # get the main target dependency
    if config.has_target():
        dep = find_dependency(root, config.target)
    else:
        dep = root

    # target init
    if config.mama_init and config.has_target():
        if not dep:
            console(f'init command failed: target {config.target} not found')
            exit(-1)
        mama_init_project(dep)
        return

    flat_deps = get_flat_deps(root) # root, dep2, deepest_dep
    flat_deps_reverse = list(reversed(flat_deps)) # deepest_dep, dep2, root

    if config.list:
        flat_deps_names = [d.name for d in flat_deps]
        if config.no_specific_target():
            console(f'    ALL Dependency List: {flat_deps_names}', Color.BLUE)
            for d in flat_deps: print_package_exports(d)
        else:
            console(f'    {dep.name} Dependency List: {flat_deps_names}', Color.BLUE)
            print_package_exports(dep)
        return

    if config.dirty:
        if not dep:
            console(f'dirty command failed: target {config.target} not found')
            exit(-1)
        mama_dirty(root, dep)
        return

    # initialize platform compiler config
    if config.android: config.android.android_home()
    if config.raspi:   config.raspi_bin()
    if config.yocto_linux: config.yocto_linux.init_default()
    if config.mips:        config.mips.init_default()

    if config.verbose:
        chain = ' -> '.join([d.name for d in flat_deps_reverse])
        console(f'Executing task chain for build:\n    {chain}', Color.BLUE)
        print_package_exports(root)

    execute_task_chain(flat_deps_reverse)

    if config.coverage_report:
        if not dep:
            console(f'coverage-report failed: target {config.target} not found')
            exit(-1)
        run_coverage_report(dep.target)
        return

    if dep and config.test and dep.get_enabled_coverage():
        console('Project was built with coverage, generating coverage report')
        run_coverage_report(dep.target)
        return

    if config.open:
        open_project(config, root)


def main(): # for backwards compat with v0.10.x
    mamabuild(sys.argv[1:])


def __main__():
    mamabuild(sys.argv[1:])


if __name__ == '__main__':
    mamabuild(sys.argv[1:])

